from scapy.layers.zigbee import ZigbeeDeviceProfile, ZigbeeAppDataPayloadStub, ZigbeeClusterLibrary, \
    _aps_profile_identifiers, _zcl_attribute_data_types, _zcl_enumerated_status_values, _DiscreteString, \
    _zcl_cluster_identifier, ZigbeeNWKStub
from scapy.fields import BitEnumField, BitField, ByteEnumField, ByteField, ConditionalField, \
    FlagsField, XBitField, XLEIntField, XLEShortField, EnumField, XShortField, PacketListField, \
    StrFixedLenField
from scapy.layers.dot15d4 import dot15d4AddressField
from scapy.packet import Packet, bind_layers

class ZigbeeZLLCommissioningCluster(Packet):
    name = "Zigbee LightLink Commissioning Cluster Frame"
    fields_desc = [
        # Frame control (8 bits)
        BitField("reserved", 0, 3),
        BitField("disable_default_response", 1, 1), # 1 not default response command will be returned
        BitEnumField("direction", 0, 1, ['client2server', 'server2client']),
        BitField("manufacturer_specific", 0, 1), # 0 manufacturer code shall not be included in the ZCL frame
        # Frame Type
        # 0b00 command acts across the entire profile
        # 0b01 command is specific to a cluster
        # 0b10 - 0b11 reserved
        BitField("zcl_frametype", 1, 2),
        # Manufacturer code (0/16 bits) only present then manufacturer_specific field is set to 1
        ConditionalField(XLEShortField("manufacturer_code", 0x0),
            lambda pkt:pkt.getfieldval("manufacturer_specific") == 1
        ),
        # Transaction sequence number (8 bits)
        ByteField("transaction_sequence", 0),
        # Command identifier (8 bits): the cluster command
        ByteEnumField("command_identifier", 0x00, {}),
    ]

class ZLLScanRequest(Packet):
    name = "ZLL: Scan Request"
    fields_desc = [
        # Inter-PAN transaction identifier (4 octets)
        XLEIntField("inter_pan_transaction_id", 0x66666666), # Unsigned 32-bit Integer (4 octets)
	# ZigBee information (1 octet)
        BitField("reserved", 0, 5),
        BitEnumField("rx_on_when_idle", 1, 1, [False, True]),
        BitEnumField("logical_type", 1, 2, {
            0:"coordinator", 1:"router", 2:"end device", 3:"reserved"}
        ),
	# ZLL information (1 octet)
        #FlagsField("ZLL information", 0, 8, [ 'factory_new', 'address_assignment', 'reserved1', 'reserved2', 'link_initiator', 'undefined', 'reserved3', 'reserved4' ]),
        BitField("reserved1", 0, 2),
        BitField("undefined", 0, 1),
        BitEnumField("link_initiator", 0, 1, [False, True]),
        BitField("reserved2", 0, 2),
        BitEnumField("address_assignment", 0, 1, [False, True]),
        BitEnumField("factory_new", 0, 1, [False, True]),
    ]
    def answers(self, other):
        if isinstance(other, ZLLScanResponse):
            return self.inter_pan_transaction_id == other.inter_pan_transaction_id
        return 0

class ZLLScanResponse(Packet):
    name = "ZLL: Scan Response"
    fields_desc = [
        # Inter-PAN transaction identifier (4 octets)
        XLEIntField("inter_pan_transaction_id", 0x66666666),
        ByteField("rssi_correction", 0x00), # range 0x00 - 0x20 (1 octet)
	# ZigBee information (1 octet)
        # HiddenField(BitField("reserved", 0, 5)),
        BitField("reserved", 0, 5),
        BitEnumField("rx_on_when_idle", 1, 1, [False, True]),
        BitEnumField("logical_type", 1, 2, {
            0:"coordinator", 1:"router", 2:"end device", 3:"reserved"}
        ),
	# ZLL information (1 octet)
        # HiddenField(BitField("reserved1", 0, 2)),
        BitField("reserved1", 0, 2),
        BitEnumField("touchlink_priority_request", 0, 1, [False, True]),
        BitEnumField("touchlink_initiator", 0, 1, [False, True]),
        # HiddenField(BitField("reserved2", 0, 2)),
        BitField("reserved2", 0, 2),
        BitEnumField("address_assignment", 0, 1, [False, True]),
        BitEnumField("factory_new", 0, 1, [False, True]),
        # Key bitmask (2 octets)
        FlagsField("key_bitmask", 0, 16, ["reserved_key_8", "reserved_key_9",
            "reserved_key_10", "reserved_key_11", "reserved_key_12",
            "reserved_key_13", "reserved_key_14", "certification_key",
            "development_key", "reserved_key_1", "reserved_key_2", "reserved_key_3",
            "master_key", "reserved_key_5", "reserved_key_6",
            "reserved_key_7"]),
        # BitField("reserved3", 0, 3),
        # BitEnumField("master_key", 0, 1, [False, True]),
        # BitField("reserved4", 0, 3),
        # BitEnumField("development_key", 0, 1, [False, True]),
        # BitEnumField("certification_key", 0, 1, [False, True]),
        # BitField("reserved5", 0, 3),
        # BitField("reserved6", 0, 4),

        # Response identifier (4 octets)
        XLEIntField("response_id", 0x66666666),
        # Extended PAN identifier (8 octets)
        dot15d4AddressField("pan_id_ext", 0, adjust=lambda pkt,x: 8),
        # Network update identifier (1 octet)
        ByteField("network_update_id", 0),
        # Logical channel (1 octet)
        ByteField("channel", 11),
        # PAN identifier (2 octets)
        XLEShortField("pan_id", 0x0000),
        # Network address (2 octets)
        XLEShortField("network_address", 0xffff),
        # Number of sub-devices (1 octet)
        ByteField("number_of_sub_devices", 1),
        # Total group identifiers (1 octet)
        ByteField("number_of_group_ids", 0),
        # Endpoint identifier (0/1 octets)
        ConditionalField(ByteField("endpoint_id", 0x00), lambda pkt:(pkt.getfieldval("number_of_sub_devices") == 1)),
        # Profile identifier (0/2 octets)
        #ConditionalField(XShortField("profile_id", 0x0000)
        ConditionalField(EnumField("profile_id", 0, _aps_profile_identifiers, fmt = "<H"), lambda pkt:(pkt.getfieldval("number_of_sub_devices") == 1)),
        # Device identifier (0/2 octets)
        ConditionalField(XShortField("device_id", 0x0000), lambda pkt:(pkt.getfieldval("number_of_sub_devices") == 1)),
        # Version (0/1 octets)
        # HiddenField(ConditionalField(BitField("0x0", 0, 4), lambda pkt:(pkt.getfieldval("number_of_sub_devices") == 1))),
        ConditionalField(BitField("unknown", 0, 4), lambda pkt:(pkt.getfieldval("number_of_sub_devices") == 1)),
        ConditionalField(BitField("application_device_version", 2, 4), lambda pkt:(pkt.getfieldval("number_of_sub_devices") == 1)),
        # Group identifier count (0/1 octets)
        ConditionalField(ByteField("group_id_count", 0x00), lambda pkt:(pkt.getfieldval("number_of_sub_devices") == 1)),
    ]

class ZLLDeviceInformationRequest(Packet):
    name = "ZLL: Device Information Request"
    fields_desc = [
        # Inter-PAN transaction identifier (4 octets)
        XLEIntField("inter_pan_transaction_id", 0x66666666),
       # Start index of device table (1 octet)
        ByteField("start_index", 0),
    ]

class ZLLDeviceInformationRecord(Packet):
    name = "ZLL: Device Information Record"
    fields_desc = [
        # Extended PAN identifier (8 octets)
        dot15d4AddressField("ieee_address", 0, adjust=lambda pkt,x: 8),
        # Endpoint identifier (1 octet)
        ByteField("endpoint_id", 0x00),
        # Profile identifier (2 octets)
        XLEShortField("profile_id", 0x0000),
        # Device identifier (2 octets)
        XLEShortField("device_id", 0x0000),
        # Version (1 octet)
        ByteField("version", 0),
        # Group identifier count (1 octet)
        ByteField("group_identifier_count", 0),
        # Sort (1 octet)
        ByteField("sort", 0),

    ]

class ZLLDeviceInformationResponse(Packet):
    name = "ZLL: Device Information Response"
    fields_desc = [
        # Inter-PAN transaction identifier (4 octets)
        XLEIntField("inter_pan_transaction_id", 0x66666666),
        # Number of sub devices (1 octet)
        ByteField("number_of_sub_devices", 0),
        # Start index of device table (1 octet)
        ByteField("start_index", 0),
        # Device information record count (1 octet)
        ByteField("device_information_record_count", 0),
        # Device Information record (variable)
        PacketListField("device_information_record", [], ZLLDeviceInformationRecord)
    ]


class ZLLIdentifyRequest(Packet):
    name = "ZLL: Identify Request"
    fields_desc = [
        # Inter-PAN transaction identifier (4 octets)
        XLEIntField("inter_pan_transaction_id", 0x66666666),
        # Identify duration (1 octet):
        #   0x0000: Exit identify mode
        #   0x0001 - 0xfffe: Number of seconds to remain in identify mode
        #   0xffff: Remain in identify mode for a default time known by the receiver
        XLEShortField("identify_duration", 0xffff),
    ]

class ZLLResetToFactoryNewRequest(Packet):
    name = "ZLL: Reset to Factory New Request"
    fields_desc = [
        # Inter-PAN transaction identifier (4 octets)
        XLEIntField("inter_pan_transaction_id", 0x66666666),
    ]

class ZLLNetworkStartRequest(Packet):
    name = "ZLL: Network Start Request"
    fields_desc = [
        # Inter-PAN transaction identifier (4 octets)
        XLEIntField("inter_pan_transaction_id", 0x66666666),
        # Extended PAN identifier (8 octets)
        dot15d4AddressField("pan_id_ext", 0, adjust=lambda pkt,x: 8),
        # Key index (1 octets)
        ByteField("key_index", 4),  # default: Master key
        # Encrypted network key (16 octets)
        StrFixedLenField("encrypted_network_key", None, length=16),#XBitField("encrypted_network_key", 0, 128),
        # Logical channel (1 octet)
        ByteField("channel", 0),
        # PAN identifier (2 octets)
        XLEShortField("pan_id", 0x0000),
        # Network address (2 octets)
        XLEShortField("network_address", 0x0001),
        # Group identifiers begin (2 octets)
        XLEShortField("group_id_begin", 0),
        # Group identifiers end (2 octets)
        XLEShortField("group_id_end", 0),
        # Free network address range begin (2 octets)
        XLEShortField("free_network_address_range_begin", 0),
        # Free network address range end (2 octets)
        XLEShortField("free_network_address_range_end", 0),
        # Free group address range begin (2 octets)
        XLEShortField("free_group_address_range_begin", 0),
        # Free group address range end (2 octets)
        XLEShortField("free_group_address_range_end", 0),
        # Initiator IEEE address (8 octet)
        XBitField("initiator_ieee_address", 0, 64),
        # Initiator network address (2 octets)
        XLEShortField("initiator_network_address", 0),
    ]

class ZLLNetworkStartResponse(Packet):
    name = "ZLL: Network Start Response"
    fields_desc = [
        # Inter-PAN transaction identifier (4 octets)
        XLEIntField("inter_pan_transaction_id", 0x66666666),
        # Status (1 octet)
        ByteEnumField("status", 0, {0: "success", 1: "failure",
            2: "reserved_status_2", 3: "reserved_status_3",
            4: "reserved_status_4", 5: "reserved_status_5",
            6: "reserved_status_6", 7: "reserved_status_7",
            8: "reserved_status_8", 9: "reserved_status_9",
            10: "reserved_status_10", 11: "reserved_status_11",
            12: "reserved_status_12", 13: "reserved_status_13",
            14: "reserved_status_14", 15: "reserved_status_15"}),
        # Extended PAN identifier (8 octets)
        dot15d4AddressField("pan_id_ext", 0, adjust=lambda pkt,x: 8),
        # Network update identifier (1 octet)
        ByteField("network_update_id", 0x00),
        # Logical channel (1 octet)
        ByteField("channel", 11),
        # PAN identifier (2 octets)
        XLEShortField("pan_id", 0x0000),
    ]

class ZLLNetworkJoinRouterRequest(Packet):
    name = "ZLL: Network Join Router Request"
    fields_desc = [
        # Inter-PAN transaction identifier (4 octets)
        XLEIntField("inter_pan_transaction_id", 0x66666666),
        # Extended PAN identifier (8 octets)
        dot15d4AddressField("pan_id_ext", 0, adjust=lambda pkt,x: 8),
        # Key index (1 octets)
        ByteField("key_index", 4),  # default: Master key
        # Encrypted network key (16 octets)
        StrFixedLenField("encrypted_network_key", None, length=16),#XBitField("encrypted_network_key", 0, 128),
        # Network update identifier (1 octet)
        ByteField("network_update_id", 0x00),
        # Logical channel (1 octet)
        ByteField("channel", 0),
        # PAN identifier (2 octets)
        XLEShortField("pan_id", 0x0000),
        # Network address (2 octets)
        XLEShortField("network_address", 0x0001),
        # Group identifiers begin (2 octets)
        XLEShortField("group_id_begin", 0),
        # Group identifiers end (2 octets)
        XLEShortField("group_id_end", 0),
        # Free network address range begin (2 octets)
        XLEShortField("free_network_address_range_begin", 0),
        # Free network address range end (2 octets)
        XLEShortField("free_network_address_range_end", 0),
        # Free group address range begin (2 octets)
        XLEShortField("free_group_address_range_begin", 0),
        # Free group address range end (2 octets)
        XLEShortField("free_group_address_range_end", 0),
    ]

class ZLLNetworkJoinRouterResponse(Packet):
    name = "ZLL: Network Join Router Response"
    fields_desc = [
        # Inter-PAN transaction identifier (4 octets)
        XLEIntField("inter_pan_transaction_id", 0x66666666),
        # Status (1 octet)
        ByteEnumField("status", 0, {0: "success", 1: "failure",
            2: "reserved_status_2", 3: "reserved_status_3",
            4: "reserved_status_4", 5: "reserved_status_5",
            6: "reserved_status_6", 7: "reserved_status_7",
            8: "reserved_status_8", 9: "reserved_status_9",
            10: "reserved_status_10", 11: "reserved_status_11",
            12: "reserved_status_12", 13: "reserved_status_13",
            14: "reserved_status_14", 15: "reserved_status_15"}),
    ]

class ZLLNetworkUpdateRequest(Packet):
    name = "ZLL: Network Update Request"
    fields_desc = [
        # Inter-PAN transaction identifier (4 octets)
        XLEIntField("inter_pan_transaction_id", 0x66666666),
        # Extended PAN identifier (8 octets)
        dot15d4AddressField("pan_id_ext", 0, adjust=lambda pkt,x: 8),
        # Network update identifier (1 octet)
        ByteField("network_update_id", 0x00),
        # Logical Channel (1 octet)
        ByteField("channel", 11),
        # PAN identifier (2 octets)
        XLEShortField("pan_id", 0x0000),
        # Network address (2 octets)
        XLEShortField("network_address", 0xffff),
    ]

def nwk_stub_guess_payload_class(self, payload):
    if self.frametype == 0b11:
        return NewZigbeeAppDataPayloadStub
    else:
        return Packet.guess_payload_class(self, payload)

ZigbeeNWKStub.guess_payload_class = nwk_stub_guess_payload_class

class NewZigbeeAppDataPayloadStub(Packet):
    name = "Zigbee Application Layer Data Payload for Inter-PAN Transmission"
    fields_desc = [
        FlagsField("frame_control", 0, 4, [ 'reserved1', 'security', 'ack_req', 'extended_hdr' ]),
        BitEnumField("delivery_mode", 0, 2, {0:'unicast', 2:'broadcast', 3:'group'}),
        BitField("frametype", 3, 2), # value 0b11 (3) is a reserved frame type
        # Group Address present only when delivery mode field has a value of 0b11 (group delivery mode)
        ConditionalField(
            XLEShortField("group_addr", 0x0), # 16-bit identifier of the group
            lambda pkt:pkt.getfieldval("delivery_mode") == 0b11
        ),
        # Cluster identifier
        EnumField("cluster", 0, _zcl_cluster_identifier, fmt = "<H"), # unsigned short (little-endian)
        # Profile identifier
        EnumField("profile", 0, _aps_profile_identifiers, fmt = "<H"),
        # ZigBee Payload
#        ConditionalField(
#            ZigbeePayloadField("data", "", length_from=lambda pkt, s:len(s)),
#            lambda pkt:pkt.frametype == 3
#        ),
    ]
    def guess_payload_class(self, payload):
        if self.frametype == 3 and self.profile == 0xc05e and self.cluster == 0x1000:
            return ZigbeeZLLCommissioningCluster
        else:
            return Packet.guess_payload_class(self, payload)


class ZCLWriteAttributeRecord(Packet):
    name = "ZCL Write Attribute Record"
    fields_desc = [
        # Attribute Identifier (2 octets)
        XLEShortField("attribute_identifier", 0),
        # Attribute Data Type (1 octet)
        ByteEnumField("attribute_data_type", 0, _zcl_attribute_data_types),
        # Attribute Data (variable)
        _DiscreteString("attribute_data", ""),
    ]

    def extract_padding(self, s):
        return "", s

class ZCLGeneralWriteAttributesUndivided(Packet):
    name = "General Domain: Command Frame Payload: write_attributes_undivided"
    fields_desc = [
        PacketListField("write_records", [], ZCLWriteAttributeRecord),
    ]

class ZCLGeneralWriteAttributesNoResponse(Packet):
    name = "General Domain: Command Frame Payload: write_attributes_no_response"
    fields_desc = [
        PacketListField("write_records", [], ZCLWriteAttributeRecord),
    ]



class ZCLConfigureReportingConfigurationRecord(Packet):
    name = "ZCL Configure Reporting Configuration Record"
    fields_desc = [
        # Direction (0/1 octet)
        ByteField("attribute_direction", 0),
        XLEShortField("attribute_identifier", 0)
    ]

    def extract_padding(self, s):
        return "", s


class ZCLGeneralReadReportingConfiguration(Packet):
    name = "General Domain: Command Frame Payload: read_reporting_configuration"
    fields_desc = [
        PacketListField("write_records", [], ZCLConfigureReportingConfigurationRecord),
    ]


class ZCLConfigureReportingConfigurationResponseRecord(Packet):
    name = "ZCL Reporting Configuration Response Record"
    fields_desc = [
        # Status (1 octet)
        ByteEnumField("status", 0, _zcl_enumerated_status_values),
        # Direction (0/1 octet)
        ConditionalField(
            ByteField("attribute_direction", 0),
            lambda pkt:pkt.status != 0x00
        ),
        # Attribute Identifier (0/2 octets)
        ConditionalField(
            XLEShortField("attribute_identifier", 0),
            lambda pkt:pkt.status != 0x00
        ),
        # Attribute Data Type (0/1 octets)
        ConditionalField(
            ByteEnumField("attribute_data_type", 0, _zcl_attribute_data_types),
            lambda pkt:pkt.status != 0x00
        ),
        # Minimum Reporting Interval (0/2 octets)
        ConditionalField(
            XLEShortField("min_reporting_interval", 0),
            lambda pkt:pkt.status != 0x00
        ),
        # Maximum Reporting Interval (0/2 octets)
        ConditionalField(
            XLEShortField("max_reporting_interval", 0),
            lambda pkt:pkt.status != 0x00
        ),
        # Reportable Change (variable)
        ConditionalField(
            _DiscreteString("reportable_change", ""),
            lambda pkt:pkt.status != 0x00
        ),
        # Timeout Period (0/2 octets)
        ConditionalField(
            XLEShortField("timeout_period", 0),
            lambda pkt:pkt.status != 0x00
        ),
    ]

    def extract_padding(self, s):
        return "", s

class ZCLGeneralReadReportingConfigurationResponse(Packet):
    name = "General Domain: Command Frame Payload: read_reporting_configuration_response"
    fields_desc = [
        PacketListField("write_records", [], ZCLConfigureReportingConfigurationResponseRecord),
    ]


class ZCLGeneralDiscoverAttributes(Packet):
    name = "General Domain: Command Frame Payload: discover_attributes"
    fields_desc = [
        XLEShortField("start_attribute_identifier", 0),
        XLEShortField("max_attribute_identifiers", 0)
    ]


class ZCLDiscoverAttributesRecord(Packet):
    name = "ZCL Discover Attributes Record"
    fields_desc = [
        # Attribute Identifier
        XLEShortField("attribute_identifier", 0),
        ByteEnumField("attribute_data_type", 0, _zcl_attribute_data_types)
    ]

    def extract_padding(self, s):
        return "", s

class ZCLGeneralDiscoverAttributesResponse(Packet):
    name = "General Domain: Command Frame Payload: discover_attributes_response"
    fields_desc = [
        ByteEnumField("discovery_complete", 0, {0: "complete", 1:"incomplete", 0xFF:"invalid"}),
        PacketListField("attribute_records", [], ZCLDiscoverAttributesRecord),
    ]

class ZCLGeneralDefaultResponse(Packet):
    name = "General Domain: Command Frame Payload: default_response"
    fields_desc = [
        ByteField("response_to_command", None),
        ByteEnumField("status", 0, _zcl_enumerated_status_values),
    ]

bind_layers(ZigbeeClusterLibrary, ZCLGeneralWriteAttributesUndivided, zcl_frametype=0x00, command_identifier=0x03)
bind_layers(ZigbeeClusterLibrary, ZCLGeneralWriteAttributesNoResponse, zcl_frametype=0x00, command_identifier=0x05)
bind_layers(ZigbeeClusterLibrary, ZCLGeneralReadReportingConfiguration, zcl_frametype=0x00, command_identifier=0x08)
bind_layers(ZigbeeClusterLibrary, ZCLGeneralReadReportingConfigurationResponse, zcl_frametype=0x00, command_identifier=0x09)
bind_layers(ZigbeeClusterLibrary, ZCLGeneralDiscoverAttributes, zcl_frametype=0x00, command_identifier=0x0c)
bind_layers(ZigbeeClusterLibrary, ZCLGeneralDiscoverAttributesResponse, zcl_frametype=0x00, command_identifier=0x0d)
bind_layers(ZigbeeClusterLibrary, ZCLGeneralDefaultResponse, zcl_frametype=0x00, command_identifier=0x0b)


bind_layers( ZigbeeZLLCommissioningCluster, ZLLScanRequest,
        command_identifier=0x00, direction=0)
bind_layers( ZigbeeZLLCommissioningCluster, ZLLScanResponse,
        command_identifier=0x01, direction=1)
bind_layers( ZigbeeZLLCommissioningCluster, ZLLDeviceInformationRequest,
        command_identifier=0x03, direction=0)
bind_layers( ZigbeeZLLCommissioningCluster, ZLLDeviceInformationResponse,
        command_identifier=0x03, direction=1)
bind_layers( ZigbeeZLLCommissioningCluster, ZLLIdentifyRequest,
        command_identifier=0x06, direction=0)
bind_layers( ZigbeeZLLCommissioningCluster, ZLLResetToFactoryNewRequest,
        command_identifier=0x07, direction=0)
bind_layers( ZigbeeZLLCommissioningCluster, ZLLNetworkStartRequest,
        command_identifier=0x10, direction=0)
bind_layers( ZigbeeZLLCommissioningCluster, ZLLNetworkStartResponse,
        command_identifier=0x11, direction=1)
bind_layers( ZigbeeZLLCommissioningCluster, ZLLNetworkJoinRouterRequest,
        command_identifier=0x12, direction=0)
bind_layers( ZigbeeZLLCommissioningCluster, ZLLNetworkJoinRouterResponse,
        command_identifier=0x13, direction=1)
bind_layers( ZigbeeZLLCommissioningCluster, ZLLNetworkUpdateRequest,
        command_identifier=0x16, direction=0)



'''
Triggers used to indicate when a packets sequence must be transmitted.
'''
from whad.exceptions import TriggerNotAssociated, InvalidTriggerPattern, WhadDeviceNotReady
from whad.helpers import scapy_packet_to_pattern
from scapy.packet import Packet

class Trigger:
    IDENTIFIER_COUNT = 0

    def __init__(self):
        self._connector = None
        self._triggered = False
        self._identifier = Trigger.IDENTIFIER_COUNT
        Trigger.IDENTIFIER_COUNT += 1

    @property
    def identifier(self):
        return self._identifier

    @property
    def triggered(self):
        return self._triggered

    @triggered.setter
    def triggered(self, triggered):
        self._triggered = triggered

    @property
    def connector(self):
        return self._connector

    @connector.setter
    def connector(self, connector):
        self._connector = connector

    def __del__(self):
        if self._connector is not None:
            self._connector.delete_sequence(self)


class ManualTrigger(Trigger):
    def __init__(self):
        super().__init__()

    def trigger(self):
        if self._connector is None:
            raise TriggerNotAssociated()
        self._connector.trigger(self)

class ConnectionEventTrigger(Trigger):
    def __init__(self, connection_event):
        self._connection_event = connection_event
        super().__init__()

    @property
    def connection_event(self):
        return self._connection_event

class ReceptionTrigger(Trigger):
    def __init__(self, pattern=None, mask=None, offset=None, packet=None, selected_fields=None, selected_layers=None):
        if pattern is None and packet is None:
            raise InvalidTriggerPattern()

        if pattern is not None and isinstance(pattern, bytes):
            self._pattern = pattern
            self._offset = offset if offset is not None else 0
            if mask is None:
                self._mask = bytes([0xFF for _ in range(len(self._pattern))])
            elif isinstance(mask, bytes) and len(mask) == len(self._pattern):
                self._mask = mask
            else:
                raise InvalidTriggerPattern()

        elif packet is not None and isinstance(packet, Packet):
            try:
                self._pattern, self._mask, self._offset = scapy_packet_to_pattern(packet, selected_fields, selected_layers)
            except:
                raise InvalidTriggerPattern()
        else:
            raise InvalidTriggerPattern()
        super().__init__()

    @property
    def pattern(self):
        return self._pattern

    @property
    def mask(self):
        return self._mask

    @property
    def offset(self):
        return self._offset



from typing import Union, Tuple

from scapy.packet import Packet

from whad.exceptions import UnsupportedDomain, UnsupportedCapability
from whad.device import Connector
from whad.helpers import message_filter
from whad.scapy.layers.phy import Phy_Packet
from whad.hub.discovery import Domain, Capability
from whad.hub.generic.cmdresult import Success, CommandResult
from whad.hub.phy import Commands, TxPower, Endianness, SupportedFreqRanges, \
    SchedulePacketResponse, PhyMetadata


from ..utils.helpers import lora_sf, lora_cr
from ..exceptions import UnsupportedFrequency, InvalidParameter, ScheduleFifoFull, \
    UnknownPhysicalLayer, UnknownPhysicalLayerFunction
from ..connector.translator import PhyMessageTranslator
from ..utils.definitions import OOKModulationScheme, ASKModulationScheme, \
    QPSKModulationScheme, BPSKModulationScheme, \
    FSKModulationScheme, GFSKModulationScheme, QFSKModulationScheme

class Phy(Connector):
    """
    Physical layer connector.

    This connector drives a Physical Layer capable device with specific WHAD messages.
    It is required by various role classes to interact with a real device and pre-process
    domain-specific messages.
    """
    translator = PhyMessageTranslator
    domain = "phy"

    def __init__(self, device=None):
        """
        Initialize the connector, open the device (if not already opened), discover
        the services (if not already discovered).
        """
        self.__ready = False
        super().__init__(device)

        # Capability cache
        self.__can_send = None
        self.__can_send_raw = None

        # Frequency cache
        self.__cached_supported_frequencies = None
        self.__cached_frequency = None

        # Address cache
        self.__address = None

        # Physical layer
        self.__physical_layer = None

        # Configurations
        self.__datarate = None
        self.__endianness = None
        self.__frequency = None
        self.__syncword = None
        self.__modulation = None

        # Open device and make sure it is compatible
        self.device.open()
        self.device.discover()

        # Initialize translator
        self.translator = PhyMessageTranslator(self.hub)

        # Check if device supports Logitech Unifying
        if not self.device.has_domain(Domain.Phy):
            raise UnsupportedDomain("PHY")
        else:
            self.__ready = True


    def format(self, packet: Union[Packet, bytes]) -> Tuple[Packet, int]:
        """
        Format a packet for PCAP export.
        """
        # Convert bytes into packet, if required
        if isinstance(packet, bytes):
            packet = Phy_Packet(packet)
        
            # Save current parameters as metadata
            packet.metadata = PhyMetadata(
                syncword=self.__syncword, frequency=self.__frequency, 
                endianness=self.__endianness, datarate=self.__datarate, 
                modulation=self.__modulation)
            
        # Process packet through WHAD's hub PHY formatter
        return self.hub.phy.format(packet)

    def close(self):
        self.stop()
        self.device.close()

    def can_use_ask(self):
        """
        Determine if the device can be configured to use Amplitude Shift Keying modulation scheme.
        """
        commands = self.device.get_domain_commands(Domain.Phy)
        return (commands & (1 << Commands.SetASKModulation)) > 0


    def can_use_fsk(self):
        """
        Determine if the device can be configured to use Frequency Shift Keying modulation scheme.
        """
        commands = self.device.get_domain_commands(Domain.Phy)
        return (commands & (1 << Commands.SetFSKModulation)) > 0


    def can_get_supported_frequencies(self):
        """
        Determine if the device can get a list of supported frequencies.
        """
        commands = self.device.get_domain_commands(Domain.Phy)
        return (commands & (1 << Commands.GetSupportedFrequencies)) > 0

    def can_set_frequency(self):
        """
        Determine if the device can select a specific frequency.
        """
        commands = self.device.get_domain_commands(Domain.Phy)
        return (commands & (1 << Commands.SetFrequency)) > 0

    def can_use_gfsk(self):
        """
        Determine if the device can be configured to use Gaussian Frequency Shift Keying modulation scheme.
        """
        commands = self.device.get_domain_commands(Domain.Phy)
        return (commands & (1 << Commands.SetGFSKModulation)) > 0

    def can_use_4fsk(self):
        """
        Determine if the device can be configured to use 4-Frequency Shift Keying modulation scheme.
        """
        commands = self.device.get_domain_commands(Domain.Phy)
        return (commands & (1 << Commands.Set4FSKModulation)) > 0

    def can_use_bpsk(self):
        """
        Determine if the device can be configured to use Binary Phase Shift Keying modulation scheme.
        """
        commands = self.device.get_domain_commands(Domain.Phy)
        return (commands & (1 << Commands.SetBPSKModulation)) > 0

    def can_use_qpsk(self):
        """
        Determine if the device can be configured to use Quadrature Phase Shift Keying modulation scheme.
        """
        commands = self.device.get_domain_commands(Domain.Phy)
        return (commands & (1 << Commands.SetQPSKModulation)) > 0

    def can_use_lora(self):
        """
        Determine if the device can be configured to use LoRa modulation scheme.
        """
        commands = self.device.get_domain_commands(Domain.Phy)
        return (commands & (1 << Commands.SetLoRaModulation)) > 0

    def can_schedule_packets(self):
        """
        Determine if the device can send scheduled packets.
        """
        commands = self.device.get_domain_commands(Domain.Phy)
        return (commands & (1 << Commands.ScheduleSend)) > 0

    def set_ask(self, on_off_keying=True):
        """
        Enable Amplitude Shift Keying modulation scheme.
        """
        if not self.can_use_ask():
            raise UnsupportedCapability("ASKModulation")

        # Create a SetAskMod message.
        msg = self.hub.phy.create_set_ask_mod(on_off_keying)

        resp = self.send_command(msg, message_filter(CommandResult))
        success = isinstance(resp, Success)
        if success:
            self.__configured_modulation = True
        return success

    def set_bfsk(self, deviation=250000):
        """
        Enable Frequency Shift Keying modulation scheme.

        Typical deviations:
        - ESB_1MBPS_PHY: 170000 Hz
        - BLE_1MBPS_PHY: 250000 Hz
        - ESB_2MBPS_PHY: 320000 Hz
        - BLE_2MBPS_PHY: 500000 Hz
        """
        if not self.can_use_fsk():
            raise UnsupportedCapability("FSKModulation")

        # Create a SetFskMod message.
        msg = self.hub.phy.create_set_fsk_mod(deviation)

        resp = self.send_command(msg, message_filter(CommandResult))
        success = isinstance(resp, Success)
        if success:
            self.__configured_modulation = True
        return success

    def set_4fsk(self, deviation=250000):
        """
        Enable 4-Frequency Shift Keying modulation scheme.
        """
        if not self.can_use_4fsk():
            raise UnsupportedCapability("4FSKModulation")

        # Create a Set4fskMod message.
        msg = self.hub.phy.create_set_4fsk_mod(deviation)

        resp = self.send_command(msg, message_filter(CommandResult))
        success = isinstance(resp, Success)
        if success:
            self.__configured_modulation = True
        return success


    def set_gfsk(self, deviation=250000):
        """
        Enable Gaussian Frequency Shift Keying modulation scheme.

        Typical deviations:
        - ESB_1MBPS_PHY: 170000 Hz
        - BLE_1MBPS_PHY: 250000 Hz
        - ESB_2MBPS_PHY: 320000 Hz
        - BLE_2MBPS_PHY: 500000 Hz
        """
        if not self.can_use_gfsk():
            raise UnsupportedCapability("GFSKModulation")

        # Create a SetGfskMod message.
        msg = self.hub.phy.create_set_gfsk_mod(deviation)

        resp = self.send_command(msg, message_filter(CommandResult))
        success = isinstance(resp, Success)
        if success:
            self.__configured_modulation = True
        return success


    def set_bpsk(self):
        """
        Enable Binary Phase Shift Keying modulation scheme.
        """
        if not self.can_use_bpsk():
            raise UnsupportedCapability("BPSKModulation")

        # Create a SetBpskMod message.
        msg = self.hub.phy.create_set_bpsk_mod()

        resp = self.send_command(msg, message_filter(CommandResult))
        success = isinstance(resp, Success)
        if success:
            self.__configured_modulation = True
        return success


    def set_qpsk(self):
        """
        Enable Quadrature Phase Shift Keying modulation scheme.
        """
        if not self.can_use_qpsk():
            raise UnsupportedCapability("QPSKModulation")


        # Create a SetQpskMod message (offset set to False by default).
        msg = self.hub.phy.create_set_qpsk_mod(False)

        resp = self.send_command(msg, message_filter(CommandResult))
        success = isinstance(resp, Success)
        if success:
            self.__configured_modulation = True
        return success


    def set_lora(self, sf=7, cr=48, bw=125000, preamble=12, crc=False, explicit=False, invert_iq=False):
        """
        Enable LoRa modulation scheme.

        @param  sf          Spreading factor (values between 7 and 12)
        @param  cr          Coding rate (values between 45 (4/5) and 48 (4/8))
        @param  bw          Bandwidth (125, 250 or 500)
        @param  preamble    Preamble length (0-65535) in number of symbols
        @param  crc         Enable CRC if set to True, disable it otherwise
        @param  explicit    LoRa explicit header mode enabled if set to True (implicit header mode if False)
        """
        if not self.can_use_lora():
            raise UnsupportedCapability("LoRaModulation")

        # Make sure parameters are valid
        if sf not in range(7, 13):
            raise InvalidParameter('spreading factor')

        if cr not in range(45, 49):
            raise InvalidParameter('coding rate')

        if bw not in [62500, 125000, 250000, 500000]:
            raise InvalidParameter('bandwidth')

        # Create a SetLoRaMod message.
        msg = self.hub.phy.create_set_lora_mod(
            bw,
            lora_sf(sf),
            lora_cr(cr),
            preamble,
            enable_crc=crc,
            explicit_mode=explicit,
            invert_iq=invert_iq
        )

        resp = self.send_command(msg, message_filter(CommandResult))
        success = isinstance(resp, Success)
        if success:
            self.__configured_modulation = True
        return success


    def get_supported_frequencies(self):
        """
        Get list of supported frequency ranges (in Hz).
        """
        if not self.can_get_supported_frequencies():
            raise UnsupportedCapability("GetSupportedFrequencies")

        # Create a GetSupportedFreqs message.
        msg = self.hub.phy.create_get_supported_freqs()

        resp = self.send_command(msg, message_filter(SupportedFreqRanges))
        assert isinstance(resp, SupportedFreqRanges)
        return [(i.start, i.end) for i in resp.ranges]


    def set_frequency(self, frequency):
        """
        Configure frequency (in Hz).
        """
        if not self.can_set_frequency():
            raise UnsupportedCapability("SetFrequency")

        if frequency is None:
            return False

        if self.__cached_supported_frequencies is None:
            self.__cached_supported_frequencies = self.get_supported_frequencies()
            #print(self.__cached_supported_frequencies)
        if all([frequency < freq_range[0] or frequency > freq_range[1] for freq_range in self.__cached_supported_frequencies]):
            raise UnsupportedFrequency(frequency)

        # Create a SetFreq message.
        msg = self.hub.phy.create_set_freq(frequency)

        resp = self.send_command(msg, message_filter(CommandResult))
        success = isinstance(resp, Success)
        if success:
            self.__cached_frequency = frequency
            self.__configured_frequency = True
        return success

    def can_set_datarate(self):
        """
        Checks if the device supports datarate configuration.
        """
        commands = self.device.get_domain_commands(Domain.Phy)
        return (commands & (1 << Commands.SetDataRate)) > 0

    def set_datarate(self, rate=1000000):
        """
        Configure datarate (in bauds).
        """

        if not self.can_set_datarate():
            raise UnsupportedCapability("DataRate")

        # Create a SetDatarate message.
        msg = self.hub.phy.create_set_datarate(rate)

        resp = self.send_command(msg, message_filter(CommandResult))
        success = isinstance(resp, Success)
        if success:
            self.__datarate = rate
        return success

    def can_set_endianness(self):
        """
        Checks if the device supports Endianness configuration.
        """
        commands = self.device.get_domain_commands(Domain.Phy)
        return (commands & (1 << Commands.SetEndianness)) > 0

    def set_endianness(self, endianness=Endianness.BIG):
        """
        Configure endianness.
        """
        if not self.can_set_endianness():
            raise UnsupportedCapability("Endianness")

        # Create a SetEndianness message.
        msg = self.hub.phy.create_set_endianness(endianness==Endianness.LITTLE)

        resp = self.send_command(msg, message_filter(CommandResult))
        success = isinstance(resp, Success)
        if success:
            self.__endianness = endianness
        return success

    def can_send(self):
        """
        Determine if the device can transmit packets.
        """
        if self.__can_send is None:
            commands = self.device.get_domain_commands(Domain.Phy)
            self.__can_send = ((commands & (1 << Commands.Send))>0 or (commands & (1 << Commands.SendRaw)))
        return self.__can_send

    def can_set_tx_power(self):
        """
        Checks if the device supports TX Power level configuration.
        """
        commands = self.device.get_domain_commands(Domain.Phy)
        return (commands & (1 << Commands.SetTXPower)) > 0


    def set_tx_power(self, tx_power = TxPower.MEDIUM):
        """
        Configure TX Power level.
        """
        if not self.can_set_tx_power():
            raise UnsupportedCapability("TXPower")

        # Create a SetTxPower message.
        msg = self.hub.phy.create_set_tx_power(tx_power)

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)


    def can_set_packet_size(self):
        """
        Checks if the device can configure packet size.
        """
        commands = self.device.get_domain_commands(Domain.Phy)
        return (commands & (1 << Commands.SetPacketSize)) > 0


    def set_packet_size(self, size=32):
        """
        Configure packet size in use.
        """
        if not self.can_set_packet_size():
            raise UnsupportedCapability("PacketSize")

        # Create a SetPacketSize message
        msg = self.hub.phy.create_set_packet_size(size)

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def can_set_sync_word(self):
        """
        Checks if the device can configure a synchronization word.
        """
        commands = self.device.get_domain_commands(Domain.Phy)
        return (commands & (1 << Commands.SetSyncWord)) > 0


    def set_sync_word(self, sync_word = b"\xAA\xAA"):
        """
        Configure synchronization word.
        """
        if not self.can_set_sync_word():
            raise UnsupportedCapability("SyncWord")

        # Create a SetSyncWord message
        msg = self.hub.phy.create_set_sync_word(sync_word)

        resp = self.send_command(msg, message_filter(CommandResult))
        success = isinstance(resp, Success)
        if success:
            self.__syncword = sync_word
        return success

    def can_sniff(self):
        """
        Determine if the device implements a sniffer mode.
        """
        commands = self.device.get_domain_commands(Domain.Phy)
        return (
            (commands & (1 << Commands.Sniff)) > 0 and
            (commands & (1 << Commands.Start))>0 and
            (commands & (1 << Commands.Stop))>0
        )


    def support_raw_iq_stream(self):
        """
        Determine if the device supports raw IQ stream.
        """
        if self.__can_send_raw is None:
            capabilities = self.device.get_domain_capability(Domain.Phy)
            self.__can_send_raw = not (capabilities & Capability.NoRawData)
        return self.__can_send_raw

    def sniff_phy(self, iq_stream = False):
        """
        Sniff Physical Layer packets.
        """
        if iq_stream and not self.support_raw_iq_stream():
            raise UnsupportedCapability("RawIQStream")

        """
        if not self.__configured_modulation:
            raise NoModulation()
        if not self.__configured_datarate:
            raise NoDatarate()
        if not self.__configured_packetsize:
            raise NoPacketSize()
        if not self.__configured_frequency:
            raise NoFrequency()
        if not self.__configured_syncword:
            raise NoSyncWord()
        if not self.__configured_endianness:
            raise NoEndianess()
        """

        # Create a SniffMode message
        msg = self.hub.phy.create_sniff_mode(iq_stream)

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def start(self):
        """
        Start currently enabled mode.
        """
        # Create a Start message
        msg = self.hub.phy.create_start()

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def stop(self):
        """
        Stop currently enabled mode.
        """
        # Create a Stop message
        msg = self.hub.phy.create_stop()

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def set_channel(self, channel):
        if self.__physical_layer is None:
            raise UnknownPhysicalLayer()

        if self.__physical_layer.channel_to_frequency is None:
            raise UnknownPhysicalLayerFunction("channel_to_frequency")

        if hasattr(self.__physical_layer.configuration, "channel"):
            self.__physical_layer.configuration.channel = channel

        return self.set_frequency(self.__physical_layer.channel_to_frequency(channel))


    def get_channel(self, channel):
        if self.__physical_layer is None:
            raise UnknownPhysicalLayer()

        if self.__physical_layer.frequency_to_channel is None:
            raise UnknownPhysicalLayerFunction("frequency_to_channel")

        if self.__cached_frequency is None:
            return None

        return self.__physical_layer.frequency_to_channel(self.__cached_frequency)


    def set_address(self, address):
        if self.__physical_layer is None:
            raise UnknownPhysicalLayer()

        if self.__physical_layer.format_address is None:
            raise UnknownPhysicalLayerFunction("format_address")

        self.__address = address
        self.translator.address = address

        if hasattr(self.__physical_layer.configuration, "address"):
            self.__physical_layer.configuration.address = address

        bytes_address = self.__physical_layer.format_address(self.__address)
        pattern = (self.__physical_layer.synchronization_word + bytes_address)
        self.translator.pattern_cropped_bytes = len(pattern) - 4
        self.translator.pattern = pattern
        self.set_sync_word(self.translator.pattern[-4:])


    def get_address(self, address):
        if self.__physical_layer is None:
            raise UnknownPhysicalLayer()

        return self.__address

    def set_configuration(self, configuration):
        if self.__physical_layer is None:
            raise UnknownPhysicalLayer()

        if hasattr(configuration, "channel"):
            self.set_channel(configuration.channel)

        if hasattr(configuration, "address"):
            self.set_address(configuration.address)

        self.__physical_layer.configuration = configuration

    def set_physical_layer(self, physical_layer):
        """
        Sets a specific physical layer.
        """
        if isinstance(physical_layer.modulation, OOKModulationScheme):
            success = self.set_ask(on_off_keying=True)
        elif isinstance(physical_layer.modulation, ASKModulationScheme):
            success = self.set_ask(on_off_keying=False)
        elif isinstance(physical_layer.modulation, QFSKModulationScheme):
            success = self.set_4fsk(deviation=physical_layer.modulation.deviation)
        elif isinstance(physical_layer.modulation, GFSKModulationScheme):
            success = self.set_gfsk(deviation=physical_layer.modulation.deviation)
        elif isinstance(physical_layer.modulation, FSKModulationScheme):
            success = self.set_fsk(deviation=physical_layer.modulation.deviation)
        elif isinstance(physical_layer.modulation, QPSKModulationScheme):
            success = self.set_qpsk()
        elif isinstance(physical_layer.modulation, BPSKModulationScheme):
            success = self.set_bpsk()
        else:
            return False

        if not success:
            return False

        success = self.set_endianness(physical_layer.endianness)
        if not success:
            return False

        success = self.set_datarate(physical_layer.datarate)
        if not success:
            return False

        success = self.set_sync_word(physical_layer.synchronization_word)
        if not success:
            return False

        success = self.set_packet_size(physical_layer.maximum_packet_size)
        if not success:
            return False

        supported_frequencies = self.get_supported_frequencies()
        start, end = physical_layer.frequency_range[0], physical_layer.frequency_range[1]
        success = False
        for i in supported_frequencies:
            if start >= i[0] and end <= i[1]:
                success = True
                break

        if not success:
            return False

        self.__physical_layer = physical_layer
        self.translator.physical_layer = self.__physical_layer
        return True

    def send(self, packet):
        """
        Send Phy packets .
        """
        if not self.can_send():
            raise UnsupportedCapability("Send")
        if isinstance(packet, bytes):
            packet = Phy_Packet(packet)

        # Generate TX metadata
        packet.metadata = PhyMetadata()
        packet.metadata.frequency = self.__cached_frequency

        return super().send_packet(packet)

    def schedule_send(self, packet, timestamp: float = 0.0) -> int:
        """Schedule a packet to be sent at a given time.
        """
        if not self.can_send():
            raise UnsupportedCapability("Send")
        if not self.can_schedule_packets():
            raise UnsupportedCapability("ScheduledSend")

        if isinstance(packet, bytes):
            packet = Phy_Packet(packet)

        # Generate TX metadata
        packet.metadata = PhyMetadata()
        packet.metadata.frequency = self.__cached_frequency

        # Set timestamp
        msg = self.hub.phy.create_schedule_packet(
            bytes(packet),
            int(timestamp*1000000)
        )

        # Schedule a packet
        resp = self.send_command(msg, message_filter(SchedulePacketResponse))
        assert isinstance(resp, SchedulePacketResponse)
        if resp.full:
            raise ScheduleFifoFull
        else:
            return resp.id

    def on_discovery_msg(self, message):
        pass

    def on_generic_msg(self, message):
        pass

    def on_domain_msg(self, domain, message):
        pass

    def on_packet(self, packet):
        """Incoming packet handler.
        """
        pass

    def on_event(self, event):
        """Incoming event handler.
        """
        pass



"""
# ATT Read procedure

The Read procedure is initiated by a Read request, and based on the required
attribute handle returns the attribute value (success) or an error (failure).
"""
from typing import List

from scapy.packet import Packet
from scapy.layers.bluetooth import ATT_Error_Response, ATT_Read_Request, ATT_Read_Response

from whad.ble.stack.att.constants import BleAttErrorCode, BleAttOpcode

from .attribute import find_attr_by_handle
from .procedure import UnexpectedProcError, BleClientProcedure, BleServerProcedure

class ServerReadProcedure(BleServerProcedure):
    """ATT server Read procedure."""

    # Read operation code
    OPCODE = BleAttOpcode.READ_REQUEST

    def __init__(self, attributes: list, mtu: int):
        """Initialize our server Read procedure."""
        super().__init__( attributes, mtu)

    @classmethod
    def trigger(cls, request) -> bool:
        """Determine if the procedure is triggered."""
        return ATT_Read_Request in request

    def process_request(self, request) -> list[Packet]:
        """React only on ReadRequest."""

        # Sanity check
        if ATT_Read_Request not in request:
            self.set_state(self.states.ERROR)
            raise UnexpectedProcError()

        # Extract Read request
        request = request[ATT_Read_Request]

        # We have a read request, look for the requested attribute
        try:
            # Query attributes
            attrib = find_attr_by_handle(self.attributes, request.gatt_handle)

            # Make sure attribute can be read
            if attrib.readable():
                # Attribute is found, return a ReadResponse and mark procedure
                # as done.
                self.set_state(self.states.DONE)
                return [ ATT_Read_Response(value=attrib.value[:self.mtu]) ]
            else:
                # Read not allowed
                self.set_state(self.states.DONE)
                return self.att_error_response(request.gatt_handle, BleAttErrorCode.READ_NOT_PERMITTED)
        except IndexError:
            self.set_state(self.states.DONE)
            return self.att_error_response(request.gatt_handle, BleAttErrorCode.ATTRIBUTE_NOT_FOUND)

class ClientReadProcedure(BleClientProcedure):
    """GATT Client Read procedure."""

    OPCODE = BleAttOpcode.READ_RESPONSE

    def __init__(self, handle: int):
        """Initialize a GATT client Read procedure.

        :param handle: Handle of attribute to read
        :type handle: int
        """
        self.__handle = handle
        super().__init__([
            ATT_Read_Request(gatt_handle=self.__handle)
        ])

    def process_request(self, request: Packet) -> List[Packet]:
        """Process incoming packet."""
        if ATT_Error_Response in request:
            self.set_result(request[ATT_Error_Response])
            self.set_state(self.states.ERROR)
        elif ATT_Read_Response in request:
            response = request[ATT_Read_Response]
            self.set_result(response.value)
            self.set_state(self.states.DONE)
        return []




from whad.dot15d4.stack.service import Dot15d4Service
from whad.rf4ce.stack.apl.exceptions import APLTimeoutException

class APLProfile(Dot15d4Service):
    """
    This class represents a APL profile, exposing a standardized API.
    """
    def __init__(self, name=None, profile_id=None):
        self.profile_id = profile_id
        super().__init__(
            manager=None,
            name=name,
            timeout_exception_class=APLTimeoutException
        )

    def attach(self, manager):
        """
        Attach this profile to the manager.
        """
        self.manager = manager

    def init(self):
        """
        Initialize the APL profile.
        """
        pass

    def on_data(self, npdu, pairing_reference, vendor_id, link_quality, rx_flags):
        """
        Callback processing incoming data for the profile.
        """
        pass

    def on_discovery(self, status, source_address, node_capability, vendor_identifier, vendor_string, application_capability, user_string, device_type_list, profile_identifier_list, requested_device_type, link_quality):
        """
        Callback processing incoming discovery request for the profile.
        """
        pass

    def on_pair(self, status, source_pan_id, source_address, node_capability, vendor_identifier, vendor_string, application_capability, user_string, device_type_list, profile_identifier_list, key_exchange_transfer_count, pairing_reference):
        """
        Callback processing incoming pairing request for the profile.
        """
        pass



"""BLE packet replay connector.

This connector can be used either in standalone to replay packets or combined
with whadreplay.
"""
from dataclasses import dataclass

from scapy.packet import Packet
from scapy.layers.bluetooth4LE import BTLE_DATA, BTLE_RF, BTLE_CTRL, LL_CONNECTION_PARAM_REQ, \
    LL_FEATURE_REQ, LL_VERSION_IND, LL_PING_REQ, LL_PING_RSP, LL_CONNECTION_PARAM_RSP, \
    LL_FEATURE_RSP
from scapy.layers.bluetooth import L2CAP_Connection_Parameter_Update_Request, \
    L2CAP_Connection_Parameter_Update_Response

from whad.device import WhadDevice
from whad.exceptions import UnsupportedCapability
from whad.scapy.layers import NordicBLE
from whad.common.replay import ReplayInterface

from .central import Central
from ..exceptions import PeripheralNotFound

@dataclass
class ReplayConfiguration:
    """
    Configuration for the Bluetooth Low Energy replay tool.

    :param target: specify the target BD address (t)
    :param random: target is using a random address (r)
    """
    target : str = None
    random : bool = True


class Replay(Central, ReplayInterface):
    """Replay BLE PDUs using a central device.

    We need to be smart here and detect any procedure initiated by a Peripheral
    to avoid sending back responses that would mess with the actual connection
    in which packets are replayed.

    This includes:

    - MTU exchange requests (ATT)
    - Feature exchange (LL_SLAVE_FEATURE_REQ, LL_FEATURE_REQ)
    - Version exchange
    - Connection parameters request procedure
    - Ping procedure
    - L2CAP Connection Update Parameter procedure
    """

    def __init__(self, device : WhadDevice, pcapfile:str, role : int):
        Central.__init__(self, device)
        ReplayInterface.__init__(self, role)
        self.__target = None
        self.__active_procedures = []

        # Make sure our device supports central mode
        if not self.can_be_central():
            raise UnsupportedCapability("Central")

    def prepare(self, config) -> bool:
        """Prepare this replay interface.
        """
        # Connect to our target device
        try:
            self.start()
            self.__target = self.connect(config.target, config.random)

            # Success !
            return True
        except PeripheralNotFound:
            # We were not able to initialize our replay instance
            return False

    def analyze_incoming_pdu(self, packet : BTLE_DATA):
        """Analyze incoming PDU to detect any procedure initiated by the remote
        Peripheral.

        Maintain the active procedures accordingly in order to avoid 
        """
        # Does the peripheral initiates a version exchange ?
        if packet.haslayer(LL_VERSION_IND) and 'version' not in self.__active_procedures:
            self.__active_procedures.append('version')
        # Or a ping ?
        elif packet.haslayer(LL_PING_REQ):
            self.__active_procedures.append('ping')
        # Or a connection parameter update procedure ?
        elif packet.haslayer(LL_CONNECTION_PARAM_REQ):
            self.__active_procedures.append('connparams')
        # Or a feature exchange procedure ?
        elif packet.haslayer(LL_FEATURE_REQ):
            self.__active_procedures.append('features')
        # Or a L2CAP connection parameter update procedure ?
        elif packet.haslayer(L2CAP_Connection_Parameter_Update_Request):
            self.__active_procedures.append('llconnparams')


    def should_send_pdu(self, packet: BTLE_DATA):
        """Determine if we should send a specific PDU
        """
        # Is this PDU an answer to a version exchange procedure initiated by the remote device ?
        if packet.haslayer(LL_VERSION_IND) and 'version' in self.__active_procedures:
            # Procedure is complete
            self.__active_procedures.remove('version')
            # But we shall not send this PDU
            return False

        if packet.haslayer(LL_PING_RSP) and 'ping' in self.__active_procedures:
            # Procedure is complete
            self.__active_procedures.remove('ping')
            return False

        if packet.haslayer(LL_CONNECTION_PARAM_RSP) and 'connparams' in self.__active_procedures:
            # Procedure is complete
            self.__active_procedures.remove('connparams')
            return False

        if packet.haslayer(LL_FEATURE_RSP):
            # Procedure is complete
            self.__active_procedures.remove('features')
            return False

        if packet.haslayer(L2CAP_Connection_Parameter_Update_Response) and 'llconnparams' in self.__active_procedures:
            # Procedure is complete
            self.__active_procedures.remove('llconnparams')
            return False

        # Check if interface supports sending control PDUs (required to send control PDUs)
        if packet.haslayer(BTLE_CTRL):
            return self.can_inject()

        # By default, send other PDUs.
        return True

    def send_packet(self, packet : Packet):
        """Send packet callback.

        Based on the current replay role, packet must be sent or not.
        """
        # First, we make sure this packet has a valid TAP
        if packet.haslayer(NordicBLE):
            nordic_layer = packet[NordicBLE]
            if nordic_layer.haslayer(BTLE_DATA):
                pdu = nordic_layer[BTLE_DATA]

                # Only allow replay of packets sent by master (central device)
                if (nordic_layer.flags & 2) and self.is_emitter():
                    if self.should_send_pdu(pdu):
                        self.send_data_pdu(pdu, conn_handle=self.__target.conn_handle)

                        # Packet sent, return True
                        return True

                    # Packet not sent
                    return False

                self.analyze_incoming_pdu(pdu)

        elif packet.haslayer(BTLE_RF):
            btle_rf = packet[BTLE_RF]
            if btle_rf.haslayer(BTLE_DATA):
                pdu = btle_rf[BTLE_DATA]

                # Only allow replay of packets sent by master (central device)
                if btle_rf.type == 0x2 and self.is_emitter():
                    if self.should_send_pdu(pdu):
                        self.send_data_pdu(pdu, conn_handle=self.__target.conn_handle)
                        return True
                    return False

                self.analyze_incoming_pdu(pdu)

        # Packet cannot be sent, return False
        return False
        


'''
This module provides the :class:`whad.phy.connector.lora.LoRa` connector that wraps PHY's
default :class:`whad.phy.connector.Phy` connector and make it easier to receive
and send LoRa packets.
'''
from whad.device import WhadDevice
from whad.exceptions import UnsupportedCapability
from whad.phy.connector import Phy
from whad.phy.exceptions import InvalidParameter


class LoRa(Phy):
    '''LoRa modulation/demodulation connector.
    '''

    SYNCWORD_M2M = b'\x12'
    SYNCWORD_LORAWAN = b'\x34'

    domain = "lora"

    def __init__(self, device: WhadDevice = None):
        '''Initialization
        '''
        # Initialize our underlying Phy connector.
        super().__init__(device)

        # Make sure LoRa is supported by the device
        if not self.can_use_lora():
            raise UnsupportedCapability('SetLoRaModulation')

        # Set LoRa default parameters
        self.__spreading_factor = 7             # SF7
        self.__coding_rate = 45                 # Coding rate 4/5
        self.__bandwidth = 125000               # 125 kHz
        self.__preamble_length = 12             # Preamble length (in symbols)
        self.__crc_enabled = False              # CRC disabled by default
        self.__explicit_mode = False            # Explicit mode is disabled by default
        self.__invert_iq = False                # Invert IQ is disabled by default
        self.__syncword = LoRa.SYNCWORD_M2M     # LoRa M2M by default


    ##
    # Getters for LoRa modulation parameters
    ##

    @property
    def sf(self):
        '''Current spreading factor.
        '''
        return self.__spreading_factor


    @sf.setter
    def sf(self, value: int):
        '''Spreading factor setter.
        '''
        if value in range(7, 13):
            self.__spreading_factor = value
        else:
            raise InvalidParameter('sf')


    @property
    def cr(self):
        '''Current coding rate.
        '''
        return self.__coding_rate


    @cr.setter
    def cr(self, value: int):
        '''Coding rate setter.
        '''
        if value in range(45, 49):
            self.__coding_rate = value
        else:
            raise InvalidParameter('cr')


    @property
    def bw(self):
        '''Current bandwidth.
        '''
        return self.__bandwidth


    @bw.setter
    def bw(self, value: int):
        '''Bandwidth setter.
        '''
        if value in [62500, 125000, 250000, 500000]:
            self.__bandwidth = value
        else:
            raise InvalidParameter('bw')


    @property
    def preamble_length(self):
        '''Current preamble size (in symbols).
        '''
        return self.__preamble_length


    @preamble_length.setter
    def preamble_length(self, value: int):
        '''Preamble length setter.
        '''
        if value in range(0, 65536):
            self.__preamble_length = value
        else:
            raise InvalidParameter('preamble')


    @property
    def crc_enabled(self):
        '''Current CRC configuration.
        '''
        return self.__crc_enabled


    def enable_crc(self, enabled: bool):
        '''Enable or disable CRC.
        '''
        self.__crc_enabled = enabled


    @property
    def explicit_mode(self):
        '''Current packet type (check if Explicit mode is set)
        '''
        return self.__explicit_mode


    def enable_explicit_mode(self, enabled: bool):
        '''Enable or disable explicit mode (variable packet length).
        '''
        self.__explicit_mode = enabled

    @property
    def invert_iq(self):
        """Current IQ inversion setting
        """
        return self.__invert_iq

    @invert_iq.setter
    def invert_iq(self, enabled : bool):
        """Set IQ inversion.
        """
        self.__invert_iq = enabled


    @property
    def syncword(self):
        '''Current configured synchronization word.
        '''
        return self.__syncword

    @syncword.setter
    def syncword(self, syncword: bytes):
        '''Set synchronization word.

        :param syncword: new synchronization word
        :type syncword: bytes
        '''
        self.__syncword = syncword



    ##
    # LoRa start/stop
    ##

    def start(self):
        '''Start device in LoRa mode (continuous RX)
        '''
        # Configure the device for LoRa modulation
        self.set_lora(
            self.__spreading_factor,
            self.__coding_rate,
            self.__bandwidth,
            self.__preamble_length,
            self.__crc_enabled,
            self.__explicit_mode,
            self.__invert_iq
        )

        # Set syncword
        self.set_sync_word(self.__syncword)

        # Start RX
        super().start()



"""Dot15d4 injection module

This module provides an Injector class for RF4CE protocol, that is used by
`winject` to inject RF4CE protocol.
"""
from typing import Generator, List

# Required by type hints
from scapy.packet import Packet

from whad.rf4ce.connector import RF4CE
from whad.rf4ce.injecting import InjectionConfiguration
from whad.exceptions import UnsupportedCapability
from whad.helpers import message_filter, is_message_type

class Injector(RF4CE):
    """
    RF4CE Injector interface for compatible WHAD device.
    """

    def __init__(self, device):
        """Initialize our injector.
        """
        RF4CE.__init__(self, device)

        self.__configuration = InjectionConfiguration()

        # Check if device can perform injection
        if not self.can_send():
            raise UnsupportedCapability("Inject")

    @property
    def configuration(self) -> InjectionConfiguration:
        """Retrieve this injector configuration.
        """
        return self.__configuration

    @configuration.setter
    def configuration(self, new_configuration: InjectionConfiguration):
        """Set the injector configuration.
        """
        self.__configuration = new_configuration

    @property
    def channel(self) -> int:
        """Retrieve the currently configured channel this injector will inject to.
        """
        return self.__configuration.channel

    @channel.setter
    def channel(self, channel: int = 11):
        """Set current channel.
        """
        self.__configuration.channel = channel

    def inject(self, packet):
        if self.__configuration.channel is not None:
            channel = self.__configuration.channel
        elif hasattr(packet, "metadata") and hasattr(packet.metadata, "channel"):
            channel = packet.metadata.channel
        else:
            channel = 11 # fallback to channel 11
        super().send(packet, channel=channel)



"""WHAD Protocol Logitech Unifying address messages abstraction layer.
"""
from whad.protocol.whad_pb2 import Message
from ..message import pb_bind, PbFieldBytes, PbMessageWrapper
from . import UnifyingDomain

@pb_bind(UnifyingDomain, 'set_node_addr', 1)
class SetNodeAddress(PbMessageWrapper):
    """Logitech unifying SetNodeAddress message
    """

    address = PbFieldBytes('unifying.set_node_addr.address')



"""WHAD Protocol Generic Debug message abstraction layer.
"""
from whad.hub.message import pb_bind, PbFieldInt, PbFieldBytes, PbMessageWrapper
from whad.hub.generic import Generic

@pb_bind(Generic, 'debug', 1)
class Debug(PbMessageWrapper):
    """Generic debug message.
    """
    level = PbFieldInt('generic.debug.level')
    msg = PbFieldBytes('generic.debug.data')



"""ProtocolHub exceptions
"""

class UnsupportedVersionException(Exception):
    """Exception raised when a ProtocolHub registry node cannot be found.
    """

    def __init__(self, name: str, version: int):
        """Create a UnsupportedVersionException object, keeps track of name and
        version.
        """
        super().__init__()
        self.__name = name
        self.__version = version

    def __repr__(self):
        """Exception string.
        """
        return f'UnsupportedVersionException(name={self.__name}, version={self.__version})'


"""
Write command procedure
"""
from typing import List, Optional, Any
from time import sleep

from scapy.packet import Packet
from scapy.layers.bluetooth import ATT_Write_Command, ATT_Error_Response

from whad.ble.stack.att.constants import BleAttOpcode

from .attribute import CharacteristicValue, find_attr_by_handle
from .procedure import BleClientProcedure, BleServerProcedure, UnexpectedProcError

class ServerWriteCommandProcedure(BleServerProcedure):
    """ATT server WriteCommand procedure."""

    OPCODE = BleAttOpcode.WRITE_COMMAND

    def __init__(self, attributes: list, mtu: int):
        """Initialize our ServerWriteCommand procedure."""
        super().__init__(attributes, mtu)

    @classmethod
    def trigger(cls, request) -> bool:
        """Determine if the procedure is triggered."""
        return ATT_Write_Command in request

    def process_request(self, request: Packet) -> list[Packet]:
        """React only on WriteRequest."""
        if ATT_Write_Command not in request:
            self.set_state(self.states.ERROR)
            raise UnexpectedProcError()

        request = request[ATT_Write_Command]

        try:
            # Find attribute
            attr = find_attr_by_handle(self.attributes,request.gatt_handle)

            # Write into attribute's value if allowed to
            if isinstance(attr, CharacteristicValue) and attr.writeable_without_resp():
                attr.value = request.data

        except IndexError:
            pass

        # We don't send anything in return
        self.set_state(self.states.DONE)
        return []

class ClientWriteCommandProcedure(BleClientProcedure):
    """GATT Client WriteCommand procedure"""

    def __init__(self, handle: int, value: bytes):
        """Initialize a client WriteCommand procedure."""
        super().__init__([
            ATT_Write_Command(
                gatt_handle=handle,
                data=value
            )
        ])
        # Set result to None
        self.set_result(None)

    def wait(self, timeout: Optional[float] = None) -> Optional[Any]:
        sleep(.3)
        if self.get_state() == self.states.INITIAL:
            self.set_result(None)
            self.set_state(self.states.DONE)
        return super().wait(timeout=timeout)

    def process_request(self, request: Packet) -> List[Packet]:
        """Process incoming PDUs."""
        if ATT_Error_Response in request:
            self.set_result(request[ATT_Error_Response])
            self.set_state(self.states.ERROR)
        return []




"""
Bluetooth Low Energy Peripheral connector
=========================================

WHAD provides a specific connector to create a BLE device, :class:`Peripheral`.
This connector implements a GATT server and hosts a GATT profile, allowing remote
BLE devices to connect to it and query its services, characteristics, and descriptors.
"""
import logging
from time import sleep
from queue import Queue, Empty
from threading import Thread, Event
from typing import Optional

from whad.hub.ble.bdaddr import BDAddress
from whad.hub.ble import Direction as BleDirection
from whad.exceptions import UnsupportedCapability

from .base import BLE
from ..stack import BleStack, Layer
from ..stack.gatt import GattServer, GattClientServer
from ..stack.att import ATTLayer
from ..stack.smp import CryptographicDatabase, Pairing
from ..profile import GenericProfile
from ..profile.device import PeripheralDevice
from ..profile.advdata import AdvDataFieldList, AdvFlagsField

# Logging
logger = logging.getLogger(__name__)

class PeripheralEventDisconnected:
    """Peripheral disconnected event.
    """
    def __init__(self, conn_handle: int):
        """Initialize this event

        :param conn_handle: Connection handle
        :type conn_handle: int
        """
        self.__conn_handle = conn_handle

    @property
    def conn_handle(self) -> int:
        """Connection handle

        :rtype: int
        """
        return self.__conn_handle


class PeripheralEventConnected:
    """Connected event
    """
    def __init__(self, conn_handle: int, local: BDAddress, remote: BDAddress):
        """Initialize event

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param local: Peripheral BD address
        :type local: :py:class:`whad.ble.BDAddress`
        :param remote: Central BD address
        :type remote: :py:class:`whad.ble.BDAddress`
        """
        self.__conn_handle = conn_handle
        self.__local = local
        self.__remote = remote

    @property
    def conn_handle(self) -> int:
        """Connection handle

        :rtype: int
        """
        return self.__conn_handle

    @property
    def local(self) -> BDAddress:
        """Peripheral BD address

        :rtype: :py:class:`whad.ble.BDAddress`
        """
        return self.__local

    @property
    def remote(self) -> BDAddress:
        """Central BD address

        :rtype: :py:class:`whad.ble.BDAddress`
        """
        return self.__remote

class PeripheralEventListener(Thread):
    """Peripheral event listener.
    """

    def __init__(self, callback=None):
        """Initialize event listener
        """
        super().__init__()
        self.__queue = Queue()
        self.__running = True
        self.__callback = callback

        # This thread is a daemon (must be terminated when main thread ends).
        self.daemon = True

    @property
    def queue(self):
        return self.__queue

    def notify(self, event):
        """Add event to notify
        """
        self.__queue.put(event)

    def stop(self):
        """Stop listener
        """
        self.__running = False

    def run(self):
        while self.__running:
            try:
                event = self.__queue.get(block=True, timeout=1.0)
                if event is not None:
                    if self.__callback is not None:
                        self.__callback(event)
            except Empty:
                pass

class Peripheral(BLE):
    """This BLE connector provides a way to create a peripheral device.

    A peripheral device exposes some services and characteristics that may
    be accessed by a central device. These services and characteristics are
    defined by a specific profile.
    """

    def __init__(self, device, existing_connection = None, profile=None, adv_data=None,
                 scan_data=None, bd_address=None, public=True, stack=BleStack, gatt=GattServer,
                 pairing=Pairing(), security_database=None):
        """Create a peripheral device.

        :param  device:     WHAD device to use as a peripheral
        :type   device:     :class:`whad.device.WhadDevice`
        :param  profile:    Device profile to use
        :type   profile:    :class:`whad.ble.profile.GenericProfile`
        :param  adv_data:   Advertisement data
        :type   adv_data:   :class:`whad.ble.profile.advdata.AdvDataFieldList`
        :param  scan_data:  Advertisement data sent in Scan Response
        :type   scan_data:  :class:`whad.ble.profile.advdata.AdvDataFieldList`
        :param  bd_address: Bluetooth Device address to use
        :type   bd_address: str
        :param  public:     Set to True to use a public Bluetooth Device address,
                            False to use a random one
        :type   public:     bool
        :param  stack:      Bluetooth Low Energy stack to use,
                            :class:`whad.ble.stack.BleStack` by default
        :param  gatt:       Bluetooth Low Energy GATT
        """
        super().__init__(device)

        # Initialize local peer and remote per info
        self.__local_peer = None
        self.__remote_peer = None
        self.__gatt_server = None
        self.__peripheral = None
        self.__central = None

        # Initialize stack
        self.__configure_stack(stack, gatt)

        self.connection = None
        self.__connected = Event()
        self.__disconnected = Event()
        self.__disconnected.set()
        self.__conn_handle = None
        self.__access_address = None

        # Initialize event listener
        self.__evt_listener = None

        # Initialize profile
        if profile is None:
            logger.info("No profile provided to this Peripheral instance, use a default one.")
            self.__profile = GenericProfile()
        else:
            logger.info("Peripheral will use the provided profile.")
            self.__profile = profile

        # Initialize security database
        if security_database is None:
            logger.info((
                "No security database provided to this Peripheral instance, "
                "use a default one."))
            self.__security_database = CryptographicDatabase()
        else:
            logger.info("Peripheral will use the provided security database.")
            self.__security_database = security_database

        # Initiate pairing parameters
        self.__pairing_parameters = pairing

        # Check if device accepts peripheral mode
        if not self.can_be_peripheral():
            logger.info("Capability MasterRole not supported by this WHAD device")
            raise UnsupportedCapability("Peripheral")

        # Set bd address if provided
        if bd_address is not None:
            logger.info("Set BD address to %s", bd_address)
            self.set_bd_address(bd_address, public=public)

        # If an existing connection is hijacked, simulate a connection
        if existing_connection is not None:
            self.on_connected(existing_connection)
        else:
            # If no advertising data has been set, initialize this peripheral
            # with default flags.
            if adv_data is None:
                adv_data = AdvDataFieldList(AdvFlagsField())

            # Enable peripheral mode
            logger.info("Enable peripheral mode with advertising data: %s", adv_data)
            self.enable_peripheral_mode(adv_data, scan_data)

    @property
    def listener(self):
        """Attached event listener
        """
        return self.__evt_listener

    @property
    def local_peer(self):
        """Local peer object
        """
        return self.__local_peer

    @property
    def remote_peer(self):
        """Remote peer object
        """
        return self.__remote_peer

    @property
    def conn_handle(self) -> int:
        """Connection handle
        """
        return self.__conn_handle

    @property
    def profile(self) -> Optional[GenericProfile]:
        """GATT Profile"""
        return self.__profile

    def __configure_stack(self, phy_layer=None, gatt_layer=None):
        """
        """
        # Save GATT and PHY layers
        if gatt_layer is not None:
            self.__gatt_layer = gatt_layer
        if phy_layer is not None:
            self.__phy_layer = phy_layer

            # Configure BLE stack to use our PHY class
            self.__stack = phy_layer(self)

        # Configure ATT layer to use our GATT class
        if self.__gatt_layer is not None:
            if issubclass(self.__gatt_layer, Layer) and self.__gatt_layer.alias == 'gatt':
                ATTLayer.add(self.__gatt_layer)

    def attach_event_listener(self, listener: PeripheralEventListener):
        """Attach an event queue to receive asynchronous notifications.

        :param  listener: Event listener
        :type   listener: PeripheralEventListener
        """
        # Save event queue
        self.__evt_listener = listener

    def notify_event(self, event):
        """Notify event
        """
        if self.__evt_listener is not None:
            self.__evt_listener.notify(event)

    def get_pairing_parameters(self):
        """Returns the provided pairing parameters, if any.
        """
        return self.__pairing_parameters

    def send_data_pdu(self, data, conn_handle=1, direction=BleDirection.SLAVE_TO_MASTER,
                      access_address=0x8e89bed6, encrypt=None) -> bool:
        """Send a PDU to the central device this peripheral device is connected to.

        Sending direction is set to ̀ BleDirection.SLAVE_TO_MASTER` as we need
        to send PDUs to a central device.

        :param  pdu:            PDU to send
        :type   pdu:            :class:`scapy.layers.bluetooth4LE.BTLE`
        :param  conn_handle:    Connection handle
        :type   conn_handle:    int
        :param  direction:      Sending direction (to master or slave)
        :type   direction:      :class:`whad.protocol.ble_pb2.BleDirection`, optional
        :param  access_address: Target access address
        :type   access_address: int, optional
        :return:                PDU transmission result.
        :rtype: bool
        """
        # TODO: access address is not really required in parameters, only conn_handle
        # We patch access address if it was set in the connection event
        if self.__access_address is not None:
            access_address = self.__access_address
        return super().send_data_pdu(data, conn_handle=conn_handle, direction=direction,
                                     access_address=access_address, encrypt=encrypt)


    def send_ctrl_pdu(self, pdu, conn_handle=1, direction=BleDirection.SLAVE_TO_MASTER,
                      access_address=0x8e89bed6, encrypt=None) -> bool:
        """Send a PDU to the central device this peripheral device is connected to.

        Sending direction is set to ̀ BleDirection.SLAVE_TO_MASTER` as we need 
        to send PDUs to a central device.

        :param  pdu:            PDU to send
        :type   pdu:            :class:`scapy.layers.bluetooth4LE.BTLE`
        :param  conn_handle:    Connection handle
        :type   conn_handle:    int
        :param  direction:      Sending direction (to master or slave)
        :type   direction:      :class:`whad.protocol.ble_pb2.BleDirection`, optional
        :param  access_address: Target access address
        :type   access_address: int, optional
        :return:                PDU transmission result.
        :rtype: bool
        """
        # TODO: access address is not really required in parameters, only conn_handle
        # We patch access address if it was set in the connection event
        if self.__access_address is not None:
            access_address = self.__access_address
        return super().send_ctrl_pdu(pdu, conn_handle=conn_handle, direction=direction,
                                     access_address=access_address, encrypt=encrypt)

    def set_profile(self, profile: GenericProfile):
        """Set peripheral profile.
        """
        self.__profile = profile

    def use_stack(self, clazz=BleStack):
        """Specify a stack class to use for BLE. By default, our own stack
        (BleStack) is used.

        :param  clazz:  BLE stack to use.
        :type   clazz:  :class:`whad.ble.stack.BleStack`
        """
        self.__stack = clazz(self)

    @property
    def smp(self):
        """Security Manager Protocol
        """
        if self.connection is not None:
            return self.connection.smp
        return None

    def pairing(self, pairing=None):
        """Trigger a pairing request with the provided parameters, if any.
        """
        if self.smp is None:
            return False
        if pairing is not None:
            self.__pairing_parameters = pairing

        if not self.smp.request_pairing(self.__pairing_parameters):
            return False

        while not self.smp.is_pairing_done():
            sleep(0.1)
            if self.smp.is_pairing_failed():
                return False

        self.smp.reset_state()
        return True

    @property
    def gatt(self):
        """Generic Attribute layer
        """
        if self.connection is not None:
            return self.connection.gatt
        return None

    @property
    def security_database(self):
        """Security Database
        """
        return self.__security_database

    def is_connected(self) -> bool:
        """Determine if the peripheral has an active connection from a
        GATT client.
        """
        return self.__connected.is_set()

    def wait_connection(self, timeout: float = None) -> bool:
        """Wait for a GATT client to connect to the peripheral. If a connection
        is already active, returns immediately.
        """
        return self.__connected.wait(timeout=timeout)

    def wait_disconnection(self, timeout: float = None) -> bool:
        """Wait for a GATT client to disconnect from the peripheral. If no connection
        is active, returns immediately.
        """
        return self.__disconnected.wait(timeout=timeout)

    ##############################
    # Incoming events
    ##############################

    def on_connected(self, connection_data):
        """A device has just connected to this peripheral.

        :param  connection_data:    Connection data
        :type   connection_data:    :class:`whad.protocol.ble_pb2.Connected`
        """
        # Make sure stack is correctly configured
        self.__configure_stack()

        # Retrieve the GATT server instance and set its profile
        logger.info("a device is now connected (connection handle: %d)",
                    connection_data.conn_handle)
        self.__local_peer = BDAddress.from_bytes(
            connection_data.advertiser,
            connection_data.adv_addr_type
        )
        self.__remote_peer = BDAddress.from_bytes(
            connection_data.initiator,
            connection_data.init_addr_type
        )

        # GATT server is now connected
        self.__connected.set()
        self.__disconnected.clear()
        self.__conn_handle = connection_data.conn_handle

        # Save access address if specified
        if connection_data.access_address != 0:
            self.__access_address = connection_data.access_address
        else:
            self.__access_address = None

        # Notify our stack that a connection has been created
        self.__stack.on_connection(
            connection_data.conn_handle,
            self.__local_peer,
            self.__remote_peer
        )

        # Notify event listener, if any
        self.notify_event(PeripheralEventConnected(
                connection_data.conn_handle,
                self.__local_peer,
                self.__remote_peer
            ))


    def on_disconnected(self, disconnection_data):
        """A device has just disconnected from this peripheral.

        :param  connection_data:    Connection data
        :type   connection_data:    :class:`whad.protocol.ble_pb2.Disconnected`
        """

        logger.info("a device has just disconnected (connection handle: %d)",
                    disconnection_data.conn_handle)
        self.__stack.on_disconnection(
            disconnection_data.conn_handle,
            disconnection_data.reason
        )

        # Notify peripheral device about this disconnection
        if self.__profile is not None:
            self.__profile.on_disconnect(disconnection_data.conn_handle)

        # We are now disconnected
        self.__connected.clear()
        self.__disconnected.set()

        # Notify event listener, if any
        self.notify_event(PeripheralEventDisconnected(
            disconnection_data.conn_handle
        ))

    def on_ctl_pdu(self, pdu):
        """This method is called whenever a control PDU is received.
        This PDU is then forwarded to the BLE stack to handle it.

        Peripheral devices act as a slave, so we only forward master to slave
        messages to the stack.

        :param  pdu:    BLE PDU
        :type   pdu:    :class:`scapy.layers.bluetooth4LE.BTLE`
        """
        if pdu.metadata.direction == BleDirection.MASTER_TO_SLAVE:
            logger.info("Control PDU comes from master, forward to peripheral")
            self.__stack.on_ctl_pdu(pdu.metadata.connection_handle, pdu)

    def on_data_pdu(self, pdu):
        """This method is called whenever a data PDU is received.
        This PDU is then forwarded to the BLE stack to handle it.

        :param  pdu:    BLE PDU
        :type   pdu:    :class:`scapy.layers.bluetooth4LE.BTLE_DATA`
        """
        if pdu.metadata.direction == BleDirection.MASTER_TO_SLAVE:
            logger.info("Data PDU comes from master, forward to peripheral")
            self.__stack.on_data_pdu(pdu.metadata.connection_handle, pdu)

    def on_new_connection(self, connection):
        """On new connection, discover primary services

        :param  connection:    Connection data
        :type   connection:    :class:`whad.protocol.ble_pb2.Connected`
        """

        # Use GATT server
        self.connection = connection

        # Retrieve GATT server
        self.__gatt_server = connection.gatt
        self.__gatt_server.set_server_model(self.__profile)
        self.__connected.set()
        self.__disconnected.clear()

        # Configure SMP layer
        # we set the security database
        self.connection.smp.set_security_database(self.__security_database)
        self.connection.smp.pairing_parameters = self.__pairing_parameters

        # Check if we got a matching LTK
        crypto_material = self.security_database.get(address=self.__local_peer)

        if crypto_material is not None and crypto_material.has_ltk():
            conn_handle = connection.conn_handle
            self.__stack.get_layer("ll").state.register_encryption_key(
                conn_handle,
                crypto_material.ltk.value
            )
            if crypto_material.is_authenticated():
                #print("Marked as authenticated")
                self.__stack.get_layer("ll").state.mark_as_authenticated(connection.conn_handle)
            else:
                pass#print("Marked as unauthenticated")


        # we indicate that we are a responder
        self.connection.smp.set_responder_role()

        # Notify our profile about this connection
        self.__profile.on_connect(self.connection.conn_handle)

    def set_mtu(self, mtu: int):
        """Set connection MTU.
        """
        if self.connection is not None:
            # Start a MTU exchange procedure
            self.connection.gatt.set_mtu(mtu)

    def get_mtu(self) -> int:
        """Retrieve the connection MTU.
        """
        if self.connection is not None:
            return self.connection.l2cap.get_local_mtu()

class PeripheralClient(Peripheral):
    '''This BLE connector provides a way to create a peripheral device with
    both GATT server and client roles.
    '''

    def __init__(self, device, existing_connection = None, profile=None, adv_data=None,
                 scan_data=None, bd_address=None, public=True, stack=BleStack):
        super().__init__(
            device,
            existing_connection=existing_connection,
            profile=profile,
            adv_data=adv_data,
            scan_data=scan_data,
            bd_address=bd_address,
            public=public,
            stack=stack
        )

        # Change ATTLayer to use GattClientServer and reinstantiate our stack
        ATTLayer.add(GattClientServer)
        self.use_stack(BleStack)


    def on_new_connection(self, connection):
        super().on_new_connection(connection)

        # Create a new peripheral device to represent the central device
        # that has just connected

        self.__peripheral = PeripheralDevice(
            self,
            connection.gatt,
            connection.conn_handle
        )

        # Retrieve GATT client
        self.__central = connection.gatt
        self.__central.set_client_model(self.__peripheral)

    @property
    def central_device(self):
        """Central device
        """
        return self.__peripheral



"""WHAD Protocol Discovery message abstraction layer.
"""
from typing import List
from whad.hub.message import HubMessage
from whad.hub.message import pb_bind, Registry
from whad.hub import ProtocolHub

class Domain:
    """
    Supported domains.
    """
    DomainNone = 0x00000000
    Phy = 0x01000000
    BtClassic = 0x02000000
    BtLE = 0x03000000
    Dot15d4 = 0x04000000
    SixLowPan = 0x05000000
    Esb = 0x06000000
    LogitechUnifying = 0x07000000
    Mosart = 0x08000000
    ANT = 0x09000000
    ANT_Plus = 0x0A000000
    ANT_FS = 0x0B000000

class Capability:
    """Device capability
    """
    CapNone = 0x00
    Scan = 0x01
    Sniff = 0x02
    Inject = 0x04
    Jam = 0x08
    Hijack = 0x10
    Hook = 0x20
    SimulateRole = 0x40
    NoRawData = 0x80

class DeviceType:
    """Device type
    """
    Esp32Blefuzzer = 0
    Butterfly = 1
    Btlejack = 2
    VirtualDevice = 4


@pb_bind(ProtocolHub, name='discovery', version=1)
class Discovery(Registry):
    """WHAD Discovery message parser/factory.
    """

    NAME = 'discovery'
    VERSIONS = {}

    def __init__(self, version: int):
        self.proto_version = version

    @staticmethod
    def parse(proto_version: int, message) -> HubMessage:
        """Parses a WHAD discovery message as seen by protobuf
        """
        message_type = message.discovery.WhichOneof('msg')
        message_clazz = Discovery.bound(message_type, proto_version)
        return message_clazz.parse(proto_version, message)

    def create_info_query(self, proto_ver: int) -> HubMessage:
        """Create a device info query message.
        """
        return Discovery.bound('info_query', self.proto_version)(
            proto_ver=proto_ver
        )

    def create_info_resp(self, type: int, device_id: bytes, proto_min_ver: int,
                              max_speed: int, fw_author: bytes, fw_url: bytes,
                              fw_version_major: int, fw_version_minor: int,
                              fw_version_rev: int, capabilities: List[int]) -> HubMessage:
        """Create a device info query response message.
        """
        return Discovery.bound('info_resp', self.proto_version)(
            type=type, device_id=device_id, proto_min_ver=proto_min_ver,
            max_speed=max_speed, fw_author=fw_author, fw_url=fw_url,
            fw_version_major=fw_version_major, fw_version_minor=fw_version_minor,
            fw_version_rev=fw_version_rev, capabilities=capabilities
        )

    def create_domain_query(self, domain: int) -> HubMessage:
        """Create a domain info query message.
        """
        return Discovery.bound('domain_query', self.proto_version)(
            domain=domain
        )

    def create_domain_resp(self, domain: int, supported_commands: int) -> HubMessage:
        """Create a device info response message
        """
        return Discovery.bound('domain_resp', self.proto_version)(
            domain=domain, supported_commands=supported_commands
        )

    def create_set_speed(self, speed: int) -> HubMessage:
        """Create a speed update message.
        """
        return Discovery.bound('set_speed', self.proto_version)(
            speed=speed
        )

    def create_reset_query(self) -> HubMessage:
        """Create a device reset query.
        """
        return Discovery.bound('reset_query', self.proto_version)()

    def create_device_ready(self) -> HubMessage:
        """Create a device ready response.
        """
        return Discovery.bound('ready_resp', self.proto_version)()

from .info import InfoQuery, InfoQueryResp
from .domain import DomainInfoQuery, DomainInfoQueryResp
from .speed import SetSpeed
from .reset import DeviceReady, ResetQuery

__all__ = [
    'Discovery',
    'InfoQuery',
    'InfoQueryResp',
    'DomainInfoQuery',
    'DomainInfoQueryResp',
    'SetSpeed',
    'DeviceReady',
    'ResetQuery'
]


"""Bluetooth Low Energy connectors.
"""
from whad.ble.connector.base import BLE
from whad.ble.connector.peripheral import Peripheral, PeripheralClient
from whad.ble.connector.central import Central
from whad.ble.connector.injector import Injector
from whad.ble.connector.hijacker import Hijacker
from whad.ble.connector.sniffer import Sniffer
from whad.ble.connector.scanner import Scanner

__all__ = [
    "BLE",
    "Peripheral",
    "PeripheralClient",
    "Central",
    "Injector",
    "Hijacker",
    "Sniffer",
    "Scanner"
]



from whad.dot15d4.connector import Dot15d4
from whad.dot15d4.stack import Dot15d4Stack
from whad.exceptions import UnsupportedCapability

class Coordinator(Dot15d4):
    """
    Zigbee Coordinator interface for compatible WHAD device.
    """
    def __init__(self, device):
        super().__init__(device)

        if not self.can_be_coordinator():
            raise UnsupportedCapability("Coordinator")

        self.__stack = Dot15d4Stack(self)
        self.__channel = 11
        self.__channel_page = 0
        self.enable_reception()

    @property
    def stack(self):
        return self.__stack

    def enable_reception(self):
        self.set_coordinator_mode(channel=self.__channel)

    def set_channel(self, channel=11):
        self.__channel = channel
        self.enable_reception()

    def get_channel(self):
        return self.__channel

    def get_channel_page(self):
        return self.__channel_page

    def perform_ed_scan(self, channel):
        if not self.can_perform_ed_scan():
            raise UnsupportedCapability("EnergyDetection")
        self.__channel = channel
        super().perform_ed_scan(channel)

    def set_channel_page(self, page=0):
        if page != 0:
            raise UnsupportedCapability("ChannelPageSelection")
        else:
            self.__channel_page = page

    def beacon(self, payload=b""):
        self.__stack.get_layer("mac").database.set("macShortAddress", 0x0000)
        self.__stack.get_layer("mac").database.set("macPanId", 0x1234)
        self.__stack.get_layer("mac").database.set("macBeaconPayload", payload)

        return self.__stack.get_layer("mac").get_service("management").beacon()

    def start_beacon(self, pan_id=0x1234):
        self.__stack.get_layer("mac").database.set("macShortAddress", 0x0000)
        self.__stack.get_layer("mac").database.set("macPanId", pan_id)
        self.__stack.get_layer("mac").database.set("macBeaconPayload", b"")

        return self.__stack.get_layer("mac").get_service("management").start(pan_id=pan_id)

    def send(self, packet):
        super().send(packet, channel=self.__channel)

    def on_pdu(self, pdu):
        if (
            hasattr(pdu,"metadata") and
            hasattr(pdu.metadata, "is_fcs_valid") and
            not pdu.metadata.is_fcs_valid
        ):
            return

        self.__stack.on_pdu(pdu)

    def on_ed_sample(self, timestamp, sample):
        self.__stack.on_ed_sample(timestamp, sample)



from whad.zigbee.stack.apl.zdo.descriptors import NodeDescriptor
from whad.dot15d4.stack.database import Dot15d4Database

class ConfigurationDatabase(Dot15d4Database):
    """
    ZigBee Device Objects configuration database.
    """
    def reset(self):
        self.configNodeDescriptor = NodeDescriptor()
        self.configNWKScanAttempts = 5
        self.configNWKTimeBetweenScans = 0xc35

        self.allowsJoins = True
        self.useWhiteList = False
        self.allowInstallCodes = False
        self.updateTrustCenterLinkKeys = True
        self.allowRejoins = False
        self.allowTrustCenterLinkKeyRequests = 0x01
        self.networkKeyUpdatePeriod = 0
        self.networkkeyUpdateMethod = 0
        self.allowApplicationKeyRequests = 1
        self.applicationKeyRequestList = []
        self.allowRemoteTcPolicyChange = False



"""
This module provides the :class:`whad.unifying.keyboard.Keyboard` class that
behaves like a Logitech Unifying keyboard. This connector is able to synchronize
with an existing keyboard dongle and send spoofed events.
"""

from whad.device import WhadDevice
from whad.unifying.connector import Unifying
from whad.esb.stack import ESBStack
from whad.unifying.stack import UnifyingApplicativeLayer, UnifyingRole, MultimediaKey
from whad.exceptions import UnsupportedCapability


class Keyboard(Unifying):
    """
    Logitech Unifying Keyboard interface for compatible WHAD device.
    """

    def __init__(self, device: WhadDevice):
        """Initialisation

        :param device: WHAD device
        :type device: WhadDevice
        """
        super().__init__(device)

        self.__channel = 5
        self.__address = "ca:e9:06:ec:a4"
        self.__started = False
        ESBStack.add(UnifyingApplicativeLayer)
        self.__stack = ESBStack(self)
        # Check if device can choose its own address
        if not self.can_set_node_address():
            raise UnsupportedCapability("SetNodeAddress")

        # Check if device can perform keyboard simulation
        if not self.can_be_keyboard():
            raise UnsupportedCapability("KeyboardSimulation")

        self._enable_role()


    def _enable_role(self):
        """Enable Unifying keyboard role.
        """
        if self.__started:
            super().stop()
        self.set_node_address(self.__address)
        self.enable_keyboard_mode(channel=self.__channel)
        self.__stack.app.role = UnifyingRole.KEYBOARD
        if self.__started:
            super().start()

    @property
    def channel(self) -> int:
        """Current channel in use

        :return: current channel in use
        :rtype: int
        """
        return self.__channel

    @channel.setter
    def channel(self, channel: int = 5):
        """Set channel

        :param channel: channel to use
        :type channel: int
        """
        self.__channel = channel
        self._enable_role()

    @property
    def key(self) -> bytes:
        """Current encryption key

        :return: encryption key
        :rtype: bytes
        """
        return self.__stack.app.key

    @key.setter
    def key(self, key: bytes):
        """Set encryption key

        :param key: encryption key to use
        :type key: bytes
        """
        self.__stack.app.key = key

    @property
    def aes_counter(self) -> int:
        """Current AES counter value

        :return: AES counter value
        :rtype: int
        """
        return self.__stack.app.aes_counter

    @aes_counter.setter
    def aes_counter(self,aes_counter: int):
        """Set AES counter value

        :param aes_counter: New AES counter value
        :type aes_counter: int
        """
        self.__stack.app.aes_counter = aes_counter

    def start(self):
        """Start keyboard mode
        """
        self.__started = True
        self._enable_role()

    def stop(self):
        """Stop keyboard mode
        """
        self.__started = False
        super().stop()

    @property
    def stack(self) -> ESBStack:
        """Current ESB stack in use

        :return: ESB stack instance
        :rtype: :class:`whad.esb.stack.ESBStack`
        """
        return self.__stack

    @property
    def address(self) -> str:
        """Current device address

        :return: current device address
        :rtype: str 
        """
        return self.__address

    @address.setter
    def address(self, address: str):
        """Set current device address

        :param address: Device address (5-byte ESB address)
        :rtype address: str
        """
        self.__address = address
        self._enable_role()

    def on_pdu(self, packet):
        """Unifying packet handler, feeds the underlying stack.
        """
        self.__stack.on_pdu(packet)

    def synchronize(self, timeout: float = 10.0) -> bool:
        """Synchronize with target device.

        :param timeout: timeout in seconds
        :type timeout: float
        :return: ``True`` if synchronization succeeded, ``False`` otherwise.
        :rtype: bool
        """
        return self.__stack.ll.synchronize(timeout=timeout)

    def lock(self):
        """Lock keyboard on current channel.
        """
        return self.__stack.app.lock_channel()

    def unlock(self):
        """Unlock keyboard from current channel.
        """
        return self.__stack.app.unlock_channel()

    def send_text(self, text: str):
        """Send a series of keystrokes.

        :param text: text to send
        :type text: str
        """
        for key in text:
            self.send_key(key)

    def send_key(self, key, ctrl: bool = False, alt: bool = False, shift: bool = False,
                 gui: bool = False) -> bool:
        """Send a keypress with modifiers.

        :param key: Key to send
        :type key: bytes
        :param ctrl: CONTROL key modifier enabled if set to ``True``
        :type ctrl: bool
        :param alt: ALT key modifier enabled if set to ``True``
        :type alt: bool
        :param shift: SHIFT key modifier enabled if set to ``True``
        :type shift: bool
        :param gui: GUI key modifier enabled if set to ``True``
        :type gui: bool
        :return: ``True`` if key has been successfully injected, ``False`` otherwise.
        :rtype: bool
        """
        if self.key is not None:
            return self.__stack.app.encrypted_keystroke(key, ctrl=ctrl, alt=alt,
                                                        shift=shift, gui=gui)
        else:
            return self.__stack.app.unencrypted_keystroke(key, ctrl=ctrl, alt=alt,
                                                          shift=shift, gui=gui)

    def volume_up(self) -> bool:
        """Send a volume up keypress.
        """
        return self.__stack.app.multimedia_keystroke(MultimediaKey.VOLUME_UP)

    def volume_down(self) -> bool:
        """Send a volume down keypress.
        """
        return self.__stack.app.multimedia_keystroke(MultimediaKey.VOLUME_DOWN)

    def volume_toggle(self) -> bool:
        """Send a volume toggle keypress.
        """
        return self.__stack.app.multimedia_keystroke(MultimediaKey.VOLUME_TOGGLE)
 


"""
This module provides the :class:`whad.unifying.connector.mouse.Mouse` class
that behaves like a Logitech Unifying mouse. This class is able to synchronize
with a specific dongle and send valid mouse and wheel moves as well as clicks.

This connector must be used with a compatible Logitech Unifying device.
"""
from whad.device import WhadDevice
from whad.unifying.connector import Unifying
from whad.esb.stack import ESBStack
from whad.unifying.stack import UnifyingApplicativeLayer, UnifyingRole, ClickType
from whad.exceptions import UnsupportedCapability


class Mouse(Unifying):
    """
    Logitech Unifying Mouse interface for compatible WHAD device.
    """

    def __init__(self, device: WhadDevice):
        """Logitech mouse initialisation

        :param device: WHAD device
        :type device: :class:`whad.device.WhadDevice`
        """
        super().__init__(device)

        self.__channel = 5
        self.__address = "ca:e9:06:ec:a4"
        self.__started = False
        ESBStack.add(UnifyingApplicativeLayer)
        self.__stack = ESBStack(self)
        # Check if device can choose its own address
        if not self.can_set_node_address():
            raise UnsupportedCapability("SetNodeAddress")

        # Check if device can perform mouse simulation
        if not self.can_be_mouse():
            raise UnsupportedCapability("MouseSimulation")

        self._enable_role()


    def lock(self):
        """Lock mouse onto current channel.
        """
        return self.__stack.app.lock_channel()

    def unlock(self):
        """Unlock mouse from current channel.
        """
        return self.__stack.app.unlock_channel()

    def _enable_role(self):
        """Enable mouse role.
        """
        if self.__started:
            super().stop()
        self.set_node_address(self.__address)
        self.enable_mouse_mode(channel=self.__channel)
        self.__stack.app.role = UnifyingRole.MOUSE
        if self.__started:
            super().start()

    @property
    def channel(self) -> int:
        """Retrieve current channel.

        :return: current channel number
        :rtype: int
        """
        return self.__channel

    @channel.setter
    def channel(self, channel: int = 5):
        """Select channel.

        :param channel: channel to select (0-100)
        :type channel: int
        """
        self.__channel = channel
        self._enable_role()

    def start(self):
        """Start mouse emulation
        """
        self.__started = True
        self._enable_role()

    def stop(self):
        """Stop mouse emulation
        """
        self.__started = False
        self.unlock()
        super().stop()

    @property
    def stack(self):
        return self.__stack

    @property
    def address(self) -> ESBStack:
        """Retrieve the underlying stack

        :return: Underlying ESB stack instance
        :rtype: :class:`whad.esb.stack.ESBStack`
        """
        return self.__address

    @address.setter
    def address(self, address: str):
        """Set mouse address

        :param address: Mouse ESB address
        :type address: str
        """
        self.__address = address
        self._enable_role()

    def on_pdu(self, packet):
        """ESB packet handler. Feed the underlying ESB stack.
        """
        self.__stack.on_pdu(packet)

    def synchronize(self, timeout: float = 10.0) -> bool:
        """Synchronize with target mouse.

        :param timeout: timeout in seconds
        :type timeout: float
        :return: ``True`` if synchronization succeeded, ``False`` otherwise
        :rtype: bool
        """
        return self.__stack.ll.synchronize(timeout=timeout)

    def move(self, x: int, y: int):
        """Send a move event to the target mouse dongle.

        :param x: Delta X in ticks
        :type x: int
        :param y: Delta Y in ticks
        :type y: int
        """
        return self.__stack.app.move_mouse(x, y)

    def left_click(self):
        """Send a left click to the target mouse dongle.
        """
        return self.__stack.app.click_mouse(type=ClickType.LEFT)

    def right_click(self):
        """Send a right click to the target mouse dongle.
        """
        return self.__stack.app.click_mouse(type=ClickType.RIGHT)

    def middle_click(self):
        """Send a middle click to the target mouse dongle.
        """
        return self.__stack.app.click_mouse(type=ClickType.MIDDLE)

    def wheel_up(self):
        """Send a wheel up event to target mouse dongle.
        """
        return self.__stack.app.wheel_mouse(x=0, y=1)

    def wheel_down(self):
        """Send a wheel down event to target mouse dongle.
        """
        return self.__stack.app.wheel_mouse(x=0, y=-1)

    def wheel_right(self):
        """Send a wheel right event to target mouse dongle.
        """
        return self.__stack.app.wheel_mouse(x=1, y=0)

    def wheel_left(self):
        """Send a wheel left event to target mouse dongle.
        """
        return self.__stack.app.wheel_mouse(x=-1, y=0)



"""
Bluetooth Low Energy Tiny Stack for unit testing
================================================

This tiny stack implements a basic GATT server and client
with a single emulated GATT profile exposing two services
and a few characteristics.
"""
from typing import Optional, List

from scapy.packet import Packet
from scapy.layers.bluetooth import L2CAP_Hdr

from .l2cap import Llcap
from .server import GattServer
from .client import GattClient

class MalformedPDUError(Exception):
    """Malformed PDU Error."""

class TinyServerStack:
    """Bluetooth Low Energy tiny stack."""

    def __init__(self, conn_handle: int):
        """Initialize BLE stack."""
        self.__conn_handle = conn_handle
        self.__l2cap = Llcap(GattServer, conn_handle)

    def on_pdu(self, packet: Packet) -> Optional[List[Packet]]:
        """Process an incoming BLE PDU and return packets to be sent back.

        :param packet: Incoming BLE packet
        :type packet: scapy.packet.Packet
        """
        # Check packet has an L2CAP header and forward to our L2CAP object
        if L2CAP_Hdr in packet:
            return self.__l2cap.on_pdu(packet[L2CAP_Hdr])
        else:
            # No L2CAP header ? PDU is malformed
            raise MalformedPDUError()




"""
Logitech Unifying packets's definition classes for Scapy.
"""
import logging

from scapy.packet import Packet, bind_layers
from scapy.fields import ByteField, XByteField, X3BytesField, IntField, \
    StrFixedLenField, ShortField, ByteEnumField, XShortField, XShortEnumField, \
    FieldLenField, StrLenField, StrField, SignedByteField, FCSField
from struct import pack

from whad.scapy.layers.esb import ESB_Payload_Hdr, SBAddressField, ESB_Ping_Request, \
    guess_payload_class_esb

logger = logging.getLogger(__name__)

class Logitech_Unifying_Hdr(Packet):
    name = "Logitech Unifying Payload"
    fields_desc = [
        XByteField("dev_index",0x00),
        XByteField("frame_type",  0x00),
        FCSField("checksum",None, fmt="B")
    ]

    def pre_dissect(self,s):
        """Try to find the longest byte sequence with a valid checksum."""
        calcCksum = 0xFF
        currentByte = 0
        cs_pos = None
        #while calcCksum+1 != s[currentByte] and currentByte < len(s) - 1:
        while currentByte < len(s) - 1:
            calcCksum = (calcCksum - s[currentByte]) & 0xFF
            currentByte += 1
            if calcCksum+1 == s[currentByte]:
                cs_pos= currentByte
        if cs_pos is None:
            return s
        return  s[:currentByte+1]


    def post_dissect(self, s):
        """Override layer post_dissect() function to reset raw packet cache.
        """
        self.raw_packet_cache = None  # Reset packet to allow post_build
        return s

    def post_build(self,p,pay):
        """Re-compute checksum if needed."""
        #if self.checksum is None:
        cksum = 0xFF
        for i in (p[:2] + pay):
            cksum = (cksum - i) & 0xFF
        cksum = (cksum + 1) & 0xFF
        #else:
        #    cksum = self.checksum
        return p[:2] + pay + pack('B', cksum)


class Logitech_Wake_Up_Payload(Packet):
    name = "Logitech Wake Up Payload"
    fields_desc = [
        XByteField("dev_index",0x00),
        StrField("unknown2", b"")
    ]


class Logitech_Encrypted_Keystroke_Payload(Packet):
    name = "Logitech Encrypted Keystroke Payload"
    fields_desc = [
        StrFixedLenField('hid_data', '\0\0\0\0\0\0\0', length=7),
        ByteField("unknown",0x00),
        IntField('aes_counter',None),
        StrFixedLenField('unused', '\0\0\0\0\0\0\0', length=7)
    ]

class Logitech_Unencrypted_Keystroke_Payload(Packet):
    name = "Logitech Unencrypted Keystroke Payload"
    fields_desc = [
        StrFixedLenField('hid_data', '\0\0\0\0\0\0\0', length=7)
    ]

class Logitech_Multimedia_Key_Payload(Packet):
    name = "Multimedia Key Payload"
    fields_desc = [
        StrFixedLenField('hid_key_scan_code', '\0\0\0\0', length=4),
        StrFixedLenField('unused','\0\0\0', length=3)
    ]

class Logitech_Keepalive_Payload(Packet):
    name = "Logitech Keepalive Payload"
    fields_desc = [
        ShortField('timeout',None)
    ]

class Logitech_Set_Keepalive_Payload(Packet):
    name = "Logitech Set Keepalive Payload"
    fields_desc = [
        ByteField("unused", None),
        ShortField('timeout',1200),
        IntField("unused_2",None)
    ]

class Logitech_Mouse_Payload(Packet):
    name = "Logitech Mouse Payload"
    fields_desc = [
        XByteField("button_mask",0x00),
        ByteField("unused",0x00),
        StrFixedLenField("movement","",length=3),
        SignedByteField("wheel_y",0x00),
        SignedByteField("wheel_x",0x00)
    ]

class Logitech_Pairing_Request_Header(Packet):
    name = "Logitech Pairing Request Header"
    fields_desc = [
        ByteEnumField("pairing_phase", None, { 1 : "first_phase", 2 : "second_phase"}),
    ]

class Logitech_Pairing_Request_1_Payload(Packet):
    name = "Logitech Pairing Request 1 Payload"
    fields_desc = [
        SBAddressField("rf_address", None, length_from=lambda _ : 5),
        ByteField("unknown1", 0x08),
        ShortField("device_wpid", None),
        ByteEnumField("protocol_id", None, {0x04: "unifying"}),
        ByteField("unknown2", None),
        XShortEnumField("device_type", None, {0x020c: "mouse", 0x0147: "keyboard"}),
        ByteField("unknown3", None),
        StrFixedLenField("padding", b"\x00\x00\x00\x00\x00", length=5)
    ]

class Logitech_Pairing_Request_2_Payload(Packet):
    name = "Logitech Pairing Request 2 Payload"
    fields_desc = [
        StrFixedLenField("device_nonce", b"\x00\x00\x00\x00", length=4),
        StrFixedLenField("device_serial", b"\x00\x00\x00\x00", length=4),
        XShortEnumField("capabilities", None, {0x0400 : "mouse", 0x1e40: "keyboard"}),
        StrFixedLenField("unknown", b"\x00\x00\x00\x00\x00\x00\x00\x00", length=8),
    ]
class Logitech_Pairing_Request_3_Payload(Packet):
    name = "Logitech Pairing Request 3 Payload"
    fields_desc = [
        ByteField("type", None),
        FieldLenField("length", None, length_of="device_name", fmt="B"),
        StrLenField("device_name",None, length_from=lambda pkt:pkt.length),
        StrField("padding", None)

    ]

class Logitech_Pairing_Response_Header(Packet):
    name = "Logitech Pairing Response Header"
    fields_desc = [
        ByteEnumField("pairing_phase", None, { 1 : "first_phase", 2 : "second_phase", 3 : "final_phase"}),
    ]

class Logitech_Pairing_Response_1_Payload(Packet):
    name = "Logitech Pairing Response 1 Payload"
    fields_desc = [
        SBAddressField("rf_address", None, length_from=lambda _ : 5),
        ByteField("unknown1", 0x08),
        ShortField("dongle_wpid", None),
        ByteEnumField("protocol_id", None, {0x04: "unifying"}),
        ByteField("unknown2", None),
        ByteField("unknown3", None),
        ByteField("unknown4", None),
        StrFixedLenField("padding", b"\x00\x00\x00\x00\x00", length=6)
    ]


class Logitech_Pairing_Response_2_Payload(Packet):
    name = "Logitech Pairing Response 2 Payload"
    fields_desc = [
        StrFixedLenField("dongle_nonce", b"\x00\x00\x00\x00", length=4),
        StrFixedLenField("dongle_serial", b"\x00\x00\x00\x00", length=4),
        XShortEnumField("capabilities", None, {0x0400 : "mouse", 0x1e40: "keyboard"}),
        StrFixedLenField("unknown", b"\x00\x00\x00\x00\x00\x00\x00\x00", length=8),
    ]

class Logitech_Pairing_Confirm_Payload(Packet):
    name = "Logitech Pairing Confirm Payload"
    fields_desc = [
        StrFixedLenField("unknown", None, length=3),
        StrFixedLenField("nonce_fragment", None, length=2),
        StrFixedLenField("serial_fragment", None, length=2),
    ]

class Logitech_Pairing_Complete_Payload(Packet):
    name = "Logitech Pairing Complete Payload"
    fields_desc = [
        StrFixedLenField("padding", b"\x00\x00\x00\x00\x00\x00\x00", length=7)
    ]


class Logitech_Waked_Up_Payload(Packet):
    name = "Logitech Waked_Up Payload"
    fields_desc = [
        XByteField("wakeup_dev_index",None),
        StrFixedLenField("unknown3", b"\x00\x1F\x00\x00\x00\xFF", length=6),

    ]

def guess_payload_class_unifying(self, payload):
    if b"\x0f\x0f\x0f\x0f" == payload[:4]:
        return ESB_Ping_Request
    elif len(payload) == 0:
        return ESB_Ack_Response
    elif len(payload) >= 2 and payload[1] in (0x51,0xC2,0x40,0x4F,0xD3,0xC1,0xC3,0x5F,0x1F, 0x0F, 0x0E, 0x10):
        return Logitech_Unifying_Hdr
    else:
        return Packet.guess_payload_class(self, payload)


# Logitech Unifying protocol
bind_layers(Logitech_Unifying_Hdr, Logitech_Waked_Up_Payload,       frame_type = 0x10)
bind_layers(Logitech_Unifying_Hdr, Logitech_Wake_Up_Payload,                frame_type = 0x51)
bind_layers(Logitech_Unifying_Hdr, Logitech_Mouse_Payload,            frame_type = 0xC2)
bind_layers(Logitech_Unifying_Hdr, Logitech_Keepalive_Payload,             frame_type = 0x40)
bind_layers(Logitech_Unifying_Hdr, Logitech_Set_Keepalive_Payload,         frame_type = 0x4F)
bind_layers(Logitech_Unifying_Hdr, Logitech_Encrypted_Keystroke_Payload,     frame_type = 0xD3)
bind_layers(Logitech_Unifying_Hdr, Logitech_Unencrypted_Keystroke_Payload,     frame_type = 0xC1)
bind_layers(Logitech_Unifying_Hdr, Logitech_Multimedia_Key_Payload,         frame_type = 0xC3)
bind_layers(Logitech_Unifying_Hdr, Logitech_Pairing_Request_Header,         frame_type = 0x5F)
bind_layers( Logitech_Pairing_Request_Header,  Logitech_Pairing_Request_1_Payload,  pairing_phase = 0x01)
bind_layers( Logitech_Pairing_Request_Header,  Logitech_Pairing_Request_2_Payload,  pairing_phase = 0x02)
bind_layers( Logitech_Pairing_Request_Header,  Logitech_Pairing_Request_3_Payload,  pairing_phase = 0x03)

bind_layers(Logitech_Unifying_Hdr, Logitech_Pairing_Response_Header,         frame_type = 0x1F)
bind_layers(Logitech_Pairing_Response_Header,  Logitech_Pairing_Response_1_Payload,  pairing_phase = 0x01)
bind_layers(Logitech_Pairing_Response_Header,  Logitech_Pairing_Response_2_Payload,  pairing_phase = 0x02)

bind_layers(Logitech_Unifying_Hdr, Logitech_Pairing_Confirm_Payload,          frame_type = 0x0F)
bind_layers(Logitech_Unifying_Hdr, Logitech_Pairing_Complete_Payload,         frame_type = 0x0E)


def bind():
    """Override the ESB_Payload_Hdr payload guessing function with ours.
    This will allow ESB PDU parsing to detect the Unifying payloads."""
    # Check if current payload class guessing function is already set to ours,
    # and update it if it is not the case.
    if ESB_Payload_Hdr.guess_payload_class != guess_payload_class_unifying:
        ESB_Payload_Hdr.guess_payload_class = guess_payload_class_unifying
        logger.debug("ESB_Payload_Hdr's payload guessing function has been updated to detect Unifying packets.")

def unbind():
    """Restore the original payload class guessing function for ESB_Payload_Hdr."""
    ESB_Payload_Hdr.guess_payload_class = guess_payload_class_esb
    logger.debug("ESB_Payload_Hdr's payload class guessing function has been restored to detect only ESB packets.")

